"""Embodies the INSTEON Product Database static data and access methods."""
import logging
import collections

from insteonplm.devices.unknowndevice import UnknownDevice
from insteonplm.devices.generalController import GeneralController
from insteonplm.devices.dimmableLightingControl import DimmableLightingControl
from insteonplm.devices.dimmableLightingControl import DimmableLightingControl_2475F
from insteonplm.devices.switchedLightingControl import SwitchedLightingControl
from insteonplm.devices.switchedLightingControl import SwitchedLightingControl_2663_222
from insteonplm.devices.securityHealthSafety import (SecurityHealthSafety,
                                   SecurityHealthSafety_2421,
                                   SecurityHealthSafety_2842_222,
                                   SecurityHealthSafety_2852_222,
                                   SecurityHealthSafety_2845_222,
                                   SecurityHealthSafety_2982_222)

from insteonplm.devices.sensorsActuators import SensorsActuators
from insteonplm.devices.sensorsActuators import SensorsActuators_2450
from insteonplm.devices.x10 import (X10OnOff, X10Dimmable, X10Sensor,
                                    X10AllUnitsOff, X10AllLightsOn,
                                    X10AllLightsOff)

# pylint: disable=line-too-long
# pylint: disable=too-few-public-methods

Product = collections.namedtuple('Product', 'cat subcat product_key description model deviceclass')


X10Product = collections.namedtuple('X10Product', 'feature deviceclass')

# flake8: noqa
class IPDB(object):
    """Embodies the INSTEON Product Database static data and access methods."""

    # pylint disable=line-too-long
    _products = [
        Product(None, None, None, 'Unknown Device', '', UnknownDevice),

        Product(0x00, None, None, 'Generic General Controller', '', GeneralController),
        Product(0x00, 0x04, None, 'ControLinc', '2430', GeneralController),
        Product(0x00, 0x05, 0x000034, 'RemoteLinc', '2440', GeneralController),
        Product(0x00, 0x06, None, 'ICON Tabletop Controller', '2830', GeneralController),
        Product(0x00, 0x08, None, 'EZBridge/EZServer', '', GeneralController),
        Product(0x00, 0x09, None, 'SignaLinc', '2442', GeneralController),
        Product(0x00, 0x0A, 0x000007, 'Poolux LCD Controller', '', GeneralController),
        Product(0x00, 0x0B, 0x000022, 'Range Extender', '2443', GeneralController),
        Product(0x00, 0x0C, 0x000028, 'IES Color Touchscreen', '', GeneralController),
        Product(0x00, 0x10, None, 'Mini Remote - 4 Scene', '2444A2WH4', GeneralController),
        Product(0x00, 0x11, None, 'Mini Remote - Switch', '2444A3', GeneralController),
        Product(0x00, 0x12, None, 'Mini Remote - 8 Scene', '2444A2WH8', GeneralController),
        Product(0x00, 0x14, None, 'Mini Remote - 4 Scene', '2342-432', GeneralController),
        Product(0x00, 0x15, None, 'Mini Remote - Switch', '2342-442', GeneralController),
        Product(0x00, 0x16, None, 'Mini Remote - 8 Scene', '2342-422', GeneralController),
        Product(0x00, 0x17, None, 'Mini Remote - 4 Scene', '2342-532', GeneralController),
        Product(0x00, 0x18, None, 'Mini Remote - 8 Scene', '2342-522', GeneralController),
        Product(0x00, 0x19, None, 'Mini Remote - Switch', '2342-542', GeneralController),
        Product(0x00, 0x1A, None, 'Mini Remote - 4 Scene', '2342-222', GeneralController),
        Product(0x00, 0x1B, None, 'Mini Remote - 8 Scene', '2342-232', GeneralController),
        Product(0x00, 0x1C, None, 'Mini Remote - Switch', '2342-242', GeneralController),
        Product(0x00, 0x1D, 0x000022, 'Range Extender', '2992-222', GeneralController),

        Product(0x01, None, None, 'Generic Dimmable Lighting Control', '', DimmableLightingControl),
        Product(0x01, 0x00, None, 'LampLinc Dimmer', '2456D3', DimmableLightingControl),
        Product(0x01, 0x01, None, 'SwitchLinc Dimmer', '2476D', DimmableLightingControl),
        Product(0x01, 0x02, None, 'In-LineLinc Dimmer', '2475D', DimmableLightingControl),
        Product(0x01, 0x03, None, 'ICON Dimmer Switch', '2876D', DimmableLightingControl),
        Product(0x01, 0x04, None, 'SwitchLinc Dimmer', '2476DH', DimmableLightingControl),
        Product(0x01, 0x05, 0x000041, 'Keypad Countdown Timer', '2484DWH8', DimmableLightingControl),
        Product(0x01, 0x06, None, 'LampLinc', '2456D2', DimmableLightingControl),
        Product(0x01, 0x07, None, 'ICON LampLinc', '2856D2', DimmableLightingControl),
        Product(0x01, 0x07, None, 'ICON LampLinc', '2856D2B', DimmableLightingControl),
        Product(0x01, 0x09, 0x000037, 'KeypadLinc Dimmer', '2486DWH6', DimmableLightingControl),
        Product(0x01, 0x0A, None, 'ICON In-Wall Controller', '', DimmableLightingControl),
        Product(0x01, 0x0B, None, 'Dimmer Module', '2632-422', DimmableLightingControl),
        Product(0x01, 0x0C, 0x00001D, 'KeypadLinc Dimmer', '2486DWH8', DimmableLightingControl),
        Product(0x01, 0x0D, None, 'SocketLinc', '2454D', DimmableLightingControl),
        Product(0x01, 0x0E, None, 'LampLinc Dimmer', '2457D2', DimmableLightingControl),
        Product(0x01, 0x0F, None, 'Dimmer Module', '2632-432', DimmableLightingControl),
        Product(0x01, 0x11, None, 'Dimmer Module', '2632-442', DimmableLightingControl),
        Product(0x01, 0x12, None, 'Dimmer Module', '2632-522', DimmableLightingControl),
        Product(0x01, 0x13, 0x000032, 'SwitchLinc Dimmer (Lixar)', '2676D-B', DimmableLightingControl),
        Product(0x01, 0x17, None, 'ToggleLinc Dimmer', '2466DW', DimmableLightingControl),
        Product(0x01, 0x18, None, 'ICON SwitchLinc Dimmer In-line Companion', '2474D', DimmableLightingControl),
        Product(0x01, 0x19, None, 'SwitchLinc Dimmer', '2476D', DimmableLightingControl),
        Product(0x01, 0x1A, None, 'In-LineLinc Dimmer', '2475D', DimmableLightingControl),
        Product(0x01, 0x1B, None, 'KeypadLinc Dimmer', '2486DWH6', DimmableLightingControl),
        Product(0x01, 0x1B, 0x00001D, 'KeypadLinc Dimmer', '2486DWH6', DimmableLightingControl),
        Product(0x01, 0x1C, None, 'KeypadLinc Dimmer', '2486DWH8', DimmableLightingControl),
        Product(0x01, 0x1D, None, 'SwitchLinc Dimmer', '2476DH', DimmableLightingControl),
        Product(0x01, 0x1E, None, 'ICON Dimmer Switch', '2876D', DimmableLightingControl),
        Product(0x01, 0x1F, 0x0000000, 'ToggleLinc Dimmer', '2466DW', DimmableLightingControl),
        Product(0x01, 0x20, None, 'SwitchLinc Dimmer', '2477D', DimmableLightingControl),
        Product(0x01, 0x20, 0x00006B, 'SwitchLinc Dimmer', '2477D', DimmableLightingControl),
        Product(0x01, 0x21, None, 'OutletLinc Dimmer', '2472DWH', DimmableLightingControl),
        Product(0x01, 0x22, None, 'LampLinc Dimmer', '2457D2X', DimmableLightingControl),
        Product(0x01, 0x23, None, 'LampLinc EZ', '2457D2', DimmableLightingControl),
        Product(0x01, 0x24, None, 'SwitchLinc 2-Wire Dimmer', '2474DWH', DimmableLightingControl),
        Product(0x01, 0x25, None, 'INSTEON Ballast Dimmer', '2475DA2', DimmableLightingControl),
        Product(0x01, 0x27, 0x000087, 'Wall Dimmer', '4701', DimmableLightingControl),
        Product(0x01, 0x29, 0x000089, 'Wall Keypad Dimmer', '4703', DimmableLightingControl),
        Product(0x01, 0x2A, 0x00008B, 'Plug-in Dimmer', '4705', DimmableLightingControl),
        Product(0x01, 0x2B, 0x000091, 'Wall Dimmer - 1000W', '4711', DimmableLightingControl),
        Product(0x01, 0x2C, 0x000092, 'In-Line Dimmer', '4712', DimmableLightingControl),
        Product(0x01, 0x2D, 0x00009E, 'SwitchLinc Dimmer', '2477DH', DimmableLightingControl),
        Product(0x01, 0x2E, None, 'FanLinc', '2475F',  DimmableLightingControl_2475F),
        Product(0x01, 0x2F, None, 'KeypadLinc Schedule Timer with Dimmer', '2484DST6', DimmableLightingControl),
        Product(0x01, 0x30, None, 'SwitchLinc Dimmer', '2476D', DimmableLightingControl),
        Product(0x01, 0x31, None, 'SwitchLinc Dimmer', '2478D', DimmableLightingControl),
        Product(0x01, 0x31, None, 'SwitchLinc Dimmer', '2478D', DimmableLightingControl),
        Product(0x01, 0x32, None, 'In-LineLinc Dimmer', '2475DA1', DimmableLightingControl),
        Product(0x01, 0x34, None, 'DIN Rail Dimmer', '2252-222', DimmableLightingControl),
        Product(0x01, 0x34, None, 'DIN Rail Dimmer', '2452-222', DimmableLightingControl),
        Product(0x01, 0x35, None, 'Micro Dimmer', '2442-222', DimmableLightingControl),
        Product(0x01, 0x36, None, 'DIN Rail Dimmer', '2452-422', DimmableLightingControl),
        Product(0x01, 0x37, None, 'DIN Rail Dimmer', '2452-522', DimmableLightingControl),
        Product(0x01, 0x38, None, 'Micro Dimmer', '2442-422', DimmableLightingControl),
        Product(0x01, 0x39, None, 'Micro Dimmer', '2442-522', DimmableLightingControl),
        Product(0x01, 0x3A, None, 'LED Bulb', '2672-222', DimmableLightingControl),
        Product(0x01, 0x3B, None, 'LED Bulb', '2672-422', DimmableLightingControl),
        Product(0x01, 0x3C, None, 'LED Bulb', '2672-522', DimmableLightingControl),
        Product(0x01, 0x3D, None, 'Ballast Dimmer', '2446-422', DimmableLightingControl),
        Product(0x01, 0x3E, None, 'Ballast Dimmer', '2446-522', DimmableLightingControl),
        Product(0x01, 0x3F, None, 'Fixture Dimmer', '2447-422', DimmableLightingControl),
        Product(0x01, 0x40, None, 'Fixture Dimmer', '2447-522', DimmableLightingControl),
        Product(0x01, 0x41, None, 'Keypad Dimmer', '2334-222', DimmableLightingControl),
        Product(0x01, 0x42, None, 'Keypad Dimmer', '2334-232', DimmableLightingControl),
        Product(0x01, 0x42, None, 'Keypad with Dimmer', '2334-232', DimmableLightingControl),
        Product(0x01, 0x49, None, 'LED PAR38 Bulb', '2674-222', DimmableLightingControl),
        Product(0x01, 0x4A, None, 'LED PAR38 Bulb', '2674-422', DimmableLightingControl),
        Product(0x01, 0x4B, None, 'LED PAR38 Bulb', '2672-522', DimmableLightingControl),
        Product(0x01, 0x4C, None, 'LED Bulb', '2672-522', DimmableLightingControl),
        Product(0x01, 0x4D, None, 'LED Bulb', '2672-522', DimmableLightingControl),
        Product(0x01, 0x4E, None, 'LED PAR38 Bulb', '2674-422', DimmableLightingControl),
        Product(0x01, 0x4F, None, 'LED PAR38 Bulb', '2672-522', DimmableLightingControl),

        Product(0x02, None, None, 'Generic Switched Lighting Control', '', SwitchedLightingControl),
        Product(0x02, 0x06, None, 'Outdoor ApplianceLinc', '2456S3E', SwitchedLightingControl),
        Product(0x02, 0x07, None, 'TimerLinc', '2456S3T', SwitchedLightingControl),
        Product(0x02, 0x08, 0x000023, 'OutletLinc Relay', '2473SWH', SwitchedLightingControl),
        Product(0x02, 0x09, None, 'ApplianceLinc', '2456S3', SwitchedLightingControl),
        Product(0x02, 0x0A, None, 'SwitchLinc Relay', '2476S', SwitchedLightingControl),
        Product(0x02, 0x0B, None, 'ICON On/Off Switch', '2876S', SwitchedLightingControl),
        Product(0x02, 0x0C, None, 'ICON Appliance Module', '2856S3B', SwitchedLightingControl),
        Product(0x02, 0x0D, None, 'ToggleLinc On/Off', '2466SW', SwitchedLightingControl),
        Product(0x02, 0x0E, None, 'SwitchLinc Relay Countdown Timer', '2476ST', SwitchedLightingControl),
        Product(0x02, 0x0F, None, 'KeypadLinc On/Off', '2486SWH6', SwitchedLightingControl),
        Product(0x02, 0x0F, 0x000036, 'KeypadLinc On/Off', '2486SWH6', SwitchedLightingControl),
        Product(0x02, 0x10, 0x00001B, 'In-LineLinc Relay', '2475S', SwitchedLightingControl),
        Product(0x02, 0x11, None, 'EZSwitch30', '', SwitchedLightingControl),
        Product(0x02, 0x12, 0x00003E, 'ICON In-LineLinc Relay', '2474S', SwitchedLightingControl),
        Product(0x02, 0x13, None, 'Icon SwitchLinc Relay (Lixar)', '2676R-B', SwitchedLightingControl),
        Product(0x02, 0x14, None, 'In-LineLinc Relay with Sense', '2475S2', SwitchedLightingControl),
        Product(0x02, 0x15, None, 'SwitchLinc Relay with Sense', '2476S', SwitchedLightingControl),
        Product(0x02, 0x16, None, 'ICON On/Off Switch', '2876S', SwitchedLightingControl),
        Product(0x02, 0x17, None, 'ICON Appliance Module', '2856S3B', SwitchedLightingControl),
        Product(0x02, 0x18, 0x000060, 'SwitchLinc 220V Relay', '2494S220', SwitchedLightingControl),
        Product(0x02, 0x19, None, 'SwitchLinc 220V Relay', '2494S220', SwitchedLightingControl),
        Product(0x02, 0x1A, 0x0000000, 'ToggleLinc On/Off', '2466SW', SwitchedLightingControl),
        Product(0x02, 0x1C, None, 'SwitchLinc Relay', '2476S', SwitchedLightingControl),
        Product(0x02, 0x1E, None, 'KeypadLinc On/Off', '2487S', SwitchedLightingControl),
        Product(0x02, 0x1F, None, 'In-LineLinc On/Off', '2475SDB', SwitchedLightingControl),
        Product(0x02, 0x20, 0x00008A, 'Wall Keypad Switch', '4704', SwitchedLightingControl),
        Product(0x02, 0x21, 0x00008C, 'Outlet Switch', '4707', SwitchedLightingControl),
        Product(0x02, 0x22, 0x000093, 'In-Line Switch', '4713', SwitchedLightingControl),
        Product(0x02, 0x23, 0x000088, 'Wall Switch', '4702', SwitchedLightingControl),
        Product(0x02, 0x24, 0x0000A1, 'Wall Keypad Switch 277V', '4102', SwitchedLightingControl),
        Product(0x02, 0x25, None, 'Keypad Countdown Timer 8-button', '2484SWH8', SwitchedLightingControl),
        Product(0x02, 0x26, None, 'KeypadLinc Schedule Timer On/Off Switch', '2485SWH6', SwitchedLightingControl),
        Product(0x02, 0x29, None, 'SwitchLinc Relay Countdown Timer', '2476ST', SwitchedLightingControl),
        Product(0x02, 0x2A, None, 'SwitchLinc Relay (Dual-Band)', '2477S', SwitchedLightingControl),
        Product(0x02, 0x2B, None, 'In-LineLinc On/Off', '2475SDB-50', SwitchedLightingControl),
        Product(0x02, 0x2B, None, 'In-LineLinc On/Off)', '2475SDB-50', SwitchedLightingControl),
        Product(0x02, 0x2C, None, 'KeypadLinc On/Off', '2487S', SwitchedLightingControl),
        Product(0x02, 0x2C, None, 'KeypadLinc On/Off', '2487S-50', SwitchedLightingControl),
        Product(0x02, 0x2D, None, 'On/Off Module', '2633-422', SwitchedLightingControl),
        Product(0x02, 0x2E, None, 'DIN Rail On/Off', '2453-222', SwitchedLightingControl),
        Product(0x02, 0x2F, None, 'Micro On/Off', '2443-222', SwitchedLightingControl),
        Product(0x02, 0x30, None, 'On/Off Module', '2633-432', SwitchedLightingControl),
        Product(0x02, 0x31, None, 'Micro On/Off', '2443-422', SwitchedLightingControl),
        Product(0x02, 0x32, None, 'Micro On/Off', '2443-522', SwitchedLightingControl),
        Product(0x02, 0x33, None, 'DIN Rail On/Off', '2453-422', SwitchedLightingControl),
        Product(0x02, 0x34, None, 'DIN Rail On/Off', '2453-522', SwitchedLightingControl),
        Product(0x02, 0x35, None, 'On/Off Module', '2633-442', SwitchedLightingControl),
        Product(0x02, 0x36, None, 'On/Off Module', '2633-522', SwitchedLightingControl),
        Product(0x02, 0x37, None, 'On/Off Module', '2635-222', SwitchedLightingControl),
        Product(0x02, 0x38, None, 'On/Off Outdoor Module', '2634-222', SwitchedLightingControl),
        Product(0x02, 0x39, None, 'On/Off Outlet', '2663-222', SwitchedLightingControl_2663_222),

        Product(0x03, None, None, 'Generic Network Bridge Controller', '', None),
        Product(0x03, 0x01, None, 'PowerLinc Serial', '2414S', None),
        Product(0x03, 0x02, None, 'PowerLinc USB', '2414U', None),
        Product(0x03, 0x03, None, 'ICON PLC Serial', '', None),
        Product(0x03, 0x04, None, 'ICON PLC USB', '', None),
        Product(0x03, 0x05, 0x00000C, 'PowerLinc Serial Modem', '2412S', None),
        Product(0x03, 0x06, None, 'IRLinc Receiver', '2411R', None),
        Product(0x03, 0x07, None, 'IRLinc Transmitter', '2411T', None),
        Product(0x03, 0x0B, 0x000030, 'PowerLinc USB Modem', '2412U', None),
        Product(0x03, 0x0D, 0x000035, 'SimpleHomeNet EZX10RF', '', None),
        Product(0x03, 0x0F, 0x00003B, 'EZX10IR', '', None),
        Product(0x03, 0x10, None, 'SmartLinc', '2412N', None),
        Product(0x03, 0x10, None, 'SmartLinc', '2412N', None),
        Product(0x03, 0x11, None, 'PowerLinc Serial Modem', '2413S', None),
        Product(0x03, 0x13, 0x000030, 'PowerLinc USB Modem', '2412UH', None),
        Product(0x03, 0x14, 0x00000C, 'PowerLinc Serial Modem', '2412SH', None),
        Product(0x03, 0x15, None, 'PowerLinc USB Modem', '2413U', None),
        Product(0x03, 0x18, None, 'Central Controller', '2243-222', None),
        Product(0x03, 0x19, None, 'PowerLinc Serial Modem', '2413SH', None),
        Product(0x03, 0x1A, None, 'PowerLinc USB Modem', '2413UH', None),
        Product(0x03, 0x1B, None, 'iGateway', '2423A4', None),
        Product(0x03, 0x1F, 0x00007E, 'USB Adapter', '2448A7', None),
        Product(0x03, 0x20, 0x00007E, 'USB Adapter', '2448A7', None),
        Product(0x03, 0x21, 0x00008E, 'USB Adapter', '2448A7H', None),
        Product(0x03, 0x22, 0x00008F, 'Central Controller Interface', '4706A', None),
        Product(0x03, 0x23, 0x00008E, 'USB Adapter', '2448A7H', None),
        Product(0x03, 0x24, 0x0000A2, 'TouchLinc', '2448A7T', None),
        Product(0x03, 0x27, 0x0000A2, 'TouchLinc', '2448A7T', None),
        Product(0x03, 0x2B, None, 'Hub', '2242-222', None),
        Product(0x03, 0x2C, None, 'Central Controller', '2243-422', None),
        Product(0x03, 0x2C, None, 'Central Controller', '2243-442', None),
        Product(0x03, 0x2D, None, 'Central Controller', '2243-522', None),
        Product(0x03, 0x2E, None, 'Hub', '2242-422', None),
        Product(0x03, 0x2F, None, 'Hub', '2242-522', None),
        Product(0x03, 0x30, None, 'Hub', '2242-442', None),
        Product(0x03, 0x31, None, 'Hub', '2242-232', None),
        Product(0x03, 0x32, None, 'Hub', '2242-222', None),
        Product(0x03, 0x33, None, 'Hub', '2245-555', None),
        Product(0x03, 0x34, None, 'Hub', '2245-442', None),
        Product(0x03, 0x35, None, 'Hub', '2245-422', None),
        Product(0x03, 0x36, None, 'Hub', '2245-522', None),
        Product(0x03, 0x37, None, 'Hub', '', None),

        Product(0x04, None, None, 'Generic Irrigation Controler', '', None),
        Product(0x04, 0x00, 0x000001, 'Compacta EZRain Sprinkler Controller', '31270', None),

        Product(0x05, None, None, 'Generic Climate Controller', '', None),
        Product(0x05, 0x00, None, 'Broan SMSC080 Exhaust Fan', '', None),
        Product(0x05, 0x01, 0x000002, 'EZTherm', '', None),
        Product(0x05, 0x02, None, 'Broan SMSC110 Exhaust Fan', '', None),
        Product(0x05, 0x03, 0x00001F, 'Thermostat Adapter', '2441V', None),
        Product(0x05, 0x04, 0x000024, 'EZTherm', '', None),
        Product(0x05, 0x05, 0x000038, 'Broan, Venmar, BEST Rangehoods', '', None),
        Product(0x05, 0x07, None, 'Wireless Thermostat', '2441ZTH', None),
        Product(0x05, 0x08, None, 'Thermostat', '2441TH', None),
        Product(0x05, 0x09, 0x000094, '7 Day Thermostat', '4715', None),
        Product(0x05, 0x0A, None, 'Wireless Thermostat', '2441ZTH', None),
        Product(0x05, 0x0B, None, 'Thermostat', '2441TH', None),
        Product(0x05, 0x0E, None, 'Integrated Remote Control Thermostat', '2491T1E', None),
        Product(0x05, 0x0F, None, 'Thermostat', '2732-422', None),
        Product(0x05, 0x10, None, 'Thermostat', '2732-522', None),
        Product(0x05, 0x11, None, 'Wireless Thermostat', '2732-432', None),
        Product(0x05, 0x12, None, 'Wireless Thermostat', '2732-532', None),
        Product(0x05, 0x13, None, 'Thermostat Heat Pump', '2732-232', None),
        Product(0x05, 0x14, None, 'Thermostat Heat Pump', '2732-432', None),
        Product(0x05, 0x15, None, 'Thermostat Heat Pump', '2732-532', None),
        Product(0x05, 0x16, None, 'Insteon Thermostat', '2441TH', None),
        Product(0x05, 0x17, None, 'Insteon Thermostat', '2732-422', None),
        Product(0x05, 0x18, None, 'Insteon Thermostat', '2732-522', None),

        Product(0x06, None, None, 'Generic Pool Controller', '', None),
        Product(0x06, 0x00, 0x000003, 'EZPool', '', None),

        Product(0x07, None, None, 'Generic Sensor Actuator', '', SensorsActuators),
        Product(0x07, 0x00, None, 'I/O Linc', '2450', SensorsActuators_2450),
        Product(0x07, 0x01, 0x000004, 'EZSns1W', '', SensorsActuators),
        Product(0x07, 0x02, 0x0000012, 'EZIO8T I/O Module', '', SensorsActuators),
        Product(0x07, 0x03, 0x0000005, 'EZIO2X4', '', SensorsActuators),
        Product(0x07, 0x04, 0x0000013, 'EZIO8SA', '', SensorsActuators),
        Product(0x07, 0x05, 0x0000014, 'EZSnsRF', '', SensorsActuators),
        Product(0x07, 0x06, 0x0000015, 'EZISnsRf', '', SensorsActuators),
        Product(0x07, 0x07, 0x0000014, 'EZIO6I', '', SensorsActuators),
        Product(0x07, 0x08, 0x0000014, 'EZIO4O', '', SensorsActuators),
        Product(0x07, 0x09, 0x0000000, 'SynchroLinc', '2423A5', SensorsActuators),
        Product(0x07, 0x0D, None, 'I/O Linc', '2450-50-60', SensorsActuators_2450),
        Product(0x07, 0x0E, None, 'I/O Module', '2248-222', SensorsActuators),
        Product(0x07, 0x0F, None, 'I/O Module', '2248-422', SensorsActuators),
        Product(0x07, 0x10, None, 'I/O Module', '2248-442', SensorsActuators),
        Product(0x07, 0x11, None, 'I/O Module', '2248-522', SensorsActuators),

        Product(0x09, None, None, 'Generic Energy Management Controller', '', None),
        Product(0x09, 0x00, 0x0000006, 'EZEnergy', '', None),
        Product(0x09, 0x01, 0x0000020, 'OnSitePro Leak Detector', '', None),
        Product(0x09, 0x02, 0x0000021, 'OnSitePro Control Valve', '', None),
        Product(0x09, 0x07, 0x0000000, 'iMeter Solo', '2423A1', None),
        Product(0x09, 0x07, 0x000006C, 'iMeter Solo', '2423A1', None),
        Product(0x09, 0x0A, 0x0000000, '220V/240V 30 AMP Load Controller Normally Open', '2477SA1', None),
        Product(0x09, 0x0B, 0x0000000, '220V/240V 30 AMP Load Controller Normally Closed', '2477SA2', None),
        Product(0x09, 0x0D, None, 'Energy Display', '2441TH', None),
        Product(0x09, 0x10, 0x0000090, 'Network Hub', '4700', None),

        Product(0x0E, None, None, 'Generic Window Coverings', '', None),
        Product(0x0E, 0x00, 0x000000B, 'Somfy Drape Controller RF Bridge', '', None),
        Product(0x0E, 0x01, 0x0000000, 'Micro Open/Close', '2444-222', None),
        Product(0x0E, 0x02, 0x0000000, 'Micro Open/Close', '2444-422', None),
        Product(0x0E, 0x03, 0x0000000, 'Micro Open/Close', '2444-522', None),

        Product(0x0F, None, None, 'Generic Plumbing Controller', '', None),
        Product(0x0F, 0x00, 0x000000E, 'Weiland Doors Central Drive and Controller', '', None),
        Product(0x0F, 0x01, 0x000000F, 'Weiland Doors Secondary Central Drive', '', None),
        Product(0x0F, 0x02, 0x0000010, 'Weiland Doors Assist Drive', '', None),
        Product(0x0F, 0x03, 0x0000011, 'Weiland Doors Elevation Drive', '', None),
        Product(0x0F, 0x04, 0x0000000, 'GarageHawk Garage Unit', '', None),
        Product(0x0F, 0x05, 0x0000000, 'GarageHawk Remote Unit', '', None),
        Product(0x0F, 0x06, 0x0000000, 'MorningLinc', '2458A1', None),
        Product(0x0F, 0x07, None, 'Deadbolt', '2863-222', None),
        Product(0x0F, 0x08, None, 'Deadbolt', '2863-422', None),
        Product(0x0F, 0x09, None, 'Deadbolt', '2863-522', None),
        Product(0x0F, 0x0A, 0x0000000, 'Lock Controller', '2862-222', None),

        Product(0x10, None, None, 'Generic Security, Heath and Safety Device', '', SecurityHealthSafety),
        Product(0x10, 0x01, None, 'Motion Sensor', '2420M', SecurityHealthSafety_2842_222),
        Product(0x10, 0x01, None, 'Motion Sensor', '2842-222', SecurityHealthSafety_2842_222),
        Product(0x10, 0x02, None, 'Open/Close Sensor', '2421', SecurityHealthSafety_2421),
        Product(0x10, 0x02, None, 'Open/Close Sensor', '2843-222', SecurityHealthSafety_2421),
        Product(0x10, 0x03, 0x0000A0, 'Motion Sensor', '4716', SecurityHealthSafety_2842_222),
        Product(0x10, 0x04, None, 'Motion Sensor', '2842-422', SecurityHealthSafety_2842_222),
        Product(0x10, 0x05, None, 'Motion Sensor', '2842-522', SecurityHealthSafety_2842_222),
        Product(0x10, 0x06, None, 'Open/Close Sensor', '2843-422', SecurityHealthSafety_2421),
        Product(0x10, 0x07, None, 'Open/Close Sensor', '2843-522', SecurityHealthSafety_2421),
        Product(0x10, 0x08, None, 'Leak Sensor', '2852-222', SecurityHealthSafety_2852_222),
        Product(0x10, 0x09, None, 'Door Sensor', '2843-232', SecurityHealthSafety_2845_222),
        Product(0x10, 0x0A, None, 'Smoke Bridge', '2982-222', SecurityHealthSafety_2982_222),
        Product(0x10, 0x11, None, 'Door Sensor', '2845-222', SecurityHealthSafety_2845_222),
        Product(0x10, 0x14, None, 'Door Sensor', '2845-422', SecurityHealthSafety_2845_222),
        Product(0x10, 0x15, None, 'Door Sensor', '2845-522', SecurityHealthSafety_2845_222),
        Product(0x10, 0x16, None, 'Motion Sensor II', '2844-222', SecurityHealthSafety_2842_222),

        Product(0xFF, 0x00, None, 'Unrecognized INSTEON Device', '', UnknownDevice),
        Product(0xFF, 0x01, None, 'Unknown Device', '', UnknownDevice),
    ]

    _x10_products = [
        X10Product("onoff", X10OnOff),
        X10Product("dimmable", X10Dimmable),
        X10Product("sensor", X10Sensor),
        X10Product("allunitsoff", X10AllUnitsOff),
        X10Product("alllightson", X10AllLightsOn),
        X10Product("alllightsoff", X10AllLightsOff)
        ]

    def __init__(self):
        """Initialize the INSTEON Product Database (IPDB)."""
        self.log = logging.getLogger(__name__)

    def __len__(self):
        """Return the length of the product database."""
        return len(self._products)  + len(self._x10_products)

    def __iter__(self):
        """Iterate through the product database."""
        for product in self._products:
            yield product

    def __getitem__(self, key):
        """Return an item from the product database."""
        cat, subcat = key

        device_product = None

        for product in self._products:
            if cat == product.cat and subcat == product.subcat:
                device_product = product

        # We failed to find a device in the database, so we will make a best
        # guess from the cat and return the generic class
        #

        if not device_product:
            for product in self._products:
                if cat == product.cat and product.subcat is None:
                    return product

        # We did not find the device or even a generic device of that category
        if not device_product:
            device_product = Product(cat, subcat, None, '', '', None)

        return device_product

    def x10(self, feature):
        """Return an X10 device based on a feature.

        Current features:
        - OnOff
        - Dimmable
        """
        x10_product = None
        for product in self._x10_products:
            if feature.lower() == product.feature:
                x10_product = product

        if not x10_product:
            x10_product = X10Product(feature, None)

        return x10_product



"""INSTEON Message User Reset."""
from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_USER_RESET_DETECTED_0X55,
                                  MESSAGE_USER_RESET_DETECTED_SIZE)


class UserReset(Message):
    """Insteon User Reset Message Received.

    Message type 0x55
    """

    _code = MESSAGE_USER_RESET_DETECTED_0X55
    _sendSize = MESSAGE_USER_RESET_DETECTED_SIZE
    _receivedSize = MESSAGE_USER_RESET_DETECTED_SIZE
    _description = 'INSTEON User Reset Message Received'

    @classmethod
    def from_raw_messsage(cls, rawmessage):
        """Create message from raw byte stream."""
        return UserReset()

    def _message_properties(self):
        return []



"""Test INSTEON Sensor Actuator devices."""

import asyncio
from insteonplm.constants import (COMMAND_LIGHT_ON_0X11_NONE,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01,
                                  MESSAGE_ACK,
                                  MESSAGE_TYPE_BROADCAST_MESSAGE,
                                  MESSAGE_TYPE_DIRECT_MESSAGE_ACK)
# SensorsActuators class not tested ?
from insteonplm.devices.sensorsActuators import SensorsActuators_2450
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from .mockPLM import MockPLM
from .mockCallbacks import MockCallbacks


def test_SensorsActuators_2450_status():
    """Test SensorActuator device model 2450."""
    def run_test(loop):
        """Asyncio test run."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'

        cat = 0x07
        subcat = 0x00
        product_key = 0x00
        description = 'I/O Linc'
        model = '2450'

        callbacks = MockCallbacks()

        device = SensorsActuators_2450(plm, address, cat, subcat,
                                       product_key, description, model)
        plm.devices[address] = device
        assert device.states[0x01].name == 'openClosedRelay'
        assert device.states[0x02].name == 'openClosedSensor'

        device.states[0x01].register_updates(callbacks.callbackmethod1)
        device.states[0x02].register_updates(callbacks.callbackmethod2)

        device.async_refresh_state()
        yield from asyncio.sleep(.1, loop=loop)

        # First state
        sentmsg = StandardSend(address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00)
        assert plm.sentmessage == sentmsg.hex

        receivedmsg = StandardSend(address,
                                   COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                                   acknak=MESSAGE_ACK)
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x55,
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK, 0))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.3, loop=loop)
        assert callbacks.callbackvalue1 == 0xff

        # Second state
        receivedmsg = StandardSend(address,
                                   COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01,
                                   acknak=MESSAGE_ACK)
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x33,
            flags=MessageFlags.create(MESSAGE_TYPE_BROADCAST_MESSAGE, 0))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)

        sentmsg = StandardSend(address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0xff
        assert callbacks.callbackvalue2 == 1

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))



"""INSTEON Message Button Event Report."""
from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_BUTTON_EVENT_REPORT_0X54,
                                  MESSAGE_BUTTON_EVENT_REPORT_SIZE)


class ButtonEventReport(Message):
    """Insteon Button Event Report Message Received.

    Message type 0x54
    """

    _code = MESSAGE_BUTTON_EVENT_REPORT_0X54
    _sendSize = MESSAGE_BUTTON_EVENT_REPORT_SIZE
    _receivedSize = MESSAGE_BUTTON_EVENT_REPORT_SIZE
    _description = 'INSTEON Standard Message Received'

    def __init__(self, event):
        """Initalize the ButtonEventReport Class."""
        self._event = event

        self._events = {0x02: 'SET button tapped',
                        0x03: 'SET button press and hold',
                        0x04: 'SET button released',
                        0x12: 'Button 2 tapped',
                        0x13: 'Button 2 press and hold',
                        0x14: 'Button 2 released',
                        0x22: 'Button 3 tapped',
                        0x23: 'Button 3 press and hold',
                        0x24: 'Button 3 released'}

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return ButtonEventReport(rawmessage[2])

    @property
    def event(self):
        """Return the button event property."""
        return self._event

    @property
    def eventText(self):
        """Return the button event text."""
        return self._events.get(self.event, None)

    def _message_properties(self):
        return [{'event': self._event}]



"""INSTEON Message X10 Send."""
from insteonplm.constants import (MESSAGE_X10_MESSAGE_SEND_0X63,
                                  MESSAGE_X10_MESSAGE_SEND_SIZE,
                                  MESSAGE_X10_MESSAGE_SEND_RECEIVED_SIZE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.messages.message import Message
import insteonplm.utils


class X10Send(Message):
    """Insteon Get Next All Link Record Message.

    Message type 0x6A
    """

    _code = MESSAGE_X10_MESSAGE_SEND_0X63
    _sendSize = MESSAGE_X10_MESSAGE_SEND_SIZE
    _receivedSize = MESSAGE_X10_MESSAGE_SEND_RECEIVED_SIZE
    _description = 'Insteon Get Next All Link Record Message'

    def __init__(self, rawX10, flag, acknak=None):
        """Initialize the X10Send Class."""
        self._rawX10 = rawX10
        self._flag = flag
        self._acknak = self._setacknak(acknak)

    @staticmethod
    def from_raw_message(rawmessage):
        """Create message from raw byte stream."""
        return X10Send(rawmessage[2], rawmessage[3], rawmessage[4:5])

    @staticmethod
    def unit_code_msg(housecode, unitcode):
        """Create an X10 message to send the house code and unit code."""
        house_byte = 0
        unit_byte = 0
        if isinstance(housecode, str):
            house_byte = insteonplm.utils.housecode_to_byte(housecode) << 4
            unit_byte = insteonplm.utils.unitcode_to_byte(unitcode)
        elif isinstance(housecode, int) and housecode < 16:
            house_byte = housecode << 4
            unit_byte = unitcode
        else:
            house_byte = housecode
            unit_byte = unitcode
        return X10Send(house_byte + unit_byte, 0x00)

    @staticmethod
    def command_msg(housecode, command):
        """Create an X10 message to send the house code and a command code."""
        house_byte = 0
        if isinstance(housecode, str):
            house_byte = insteonplm.utils.housecode_to_byte(housecode) << 4
        elif isinstance(housecode, int) and housecode < 16:
            house_byte = housecode << 4
        else:
            house_byte = housecode
        return X10Send(house_byte + command, 0x80)

    @property
    def rawX10(self):
        """Return the raw x10 bytes."""
        return self._rawX10

    @property
    def flag(self):
        """Return the X10 flag."""
        return self._flag

    @property
    def acknak(self):
        """Return the ACK/NAK byte."""
        return self._acknak

    @property
    def isack(self):
        """Test if this is an ACK message."""
        return self._acknak is not None and self._acknak == MESSAGE_ACK

    @property
    def isnak(self):
        """Test if this is a NAK message."""
        return self._acknak is not None and self._acknak == MESSAGE_NAK

    def _message_properties(self):
        return [{'rawX10': self._rawX10},
                {'flag': self._flag},
                {'acknak': self._acknak}]



"""INSTEON Message All-Link Complete."""

from insteonplm.address import Address
from insteonplm.constants import (MESSAGE_ALL_LINKING_COMPLETED_0X53,
                                  MESSAGE_ALL_LINKING_COMPLETED_SIZE)
from insteonplm.messages.message import Message


class AllLinkComplete(Message):
    """INSTEON ALL-Linking Completed Message.

    Message type 0x53
    """

    _code = MESSAGE_ALL_LINKING_COMPLETED_0X53
    _sendSize = MESSAGE_ALL_LINKING_COMPLETED_SIZE
    _receivedSize = MESSAGE_ALL_LINKING_COMPLETED_SIZE
    _description = 'INSTEON ALL-Linking Completed Message Received'

    def __init__(self, linkcode, group, address, cat, subcat, firmware):
        """Initialize the AllLinkComplete Class."""
        self._linkcode = linkcode
        self._group = group
        self._address = Address(address)
        self._category = cat
        self._subcategory = subcat
        self._firmware = firmware

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create a message from a raw byte stream."""
        return AllLinkComplete(rawmessage[2],
                               rawmessage[3],
                               rawmessage[4:7],
                               rawmessage[7],
                               rawmessage[8],
                               rawmessage[9])

    @property
    def linkcode(self):
        """Return link record link code."""
        return self._linkcode

    @property
    def group(self):
        """Return link record group."""
        return self._group

    @property
    def address(self):
        """Return the device address."""
        return self._address

    @property
    def category(self):
        """Return the device category."""
        return self._category

    @property
    def subcategory(self):
        """Return the device subcategory."""
        return self._subcategory

    @property
    def firmware(self):
        """Return the device firmware version."""
        return self._firmware

    @property
    def isresponder(self):
        """Return if the link record is a responder."""
        if self.linkcode == 0:
            return True
        else:
            return False

    @property
    def iscontroller(self):
        """Return if the link record is a controller."""
        if self.linkcode == 1:
            return True
        else:
            return False

    @property
    def isdeleted(self):
        """Return if the link record is deleted."""
        if self.linkcode == 0xFF:
            return True
        else:
            return False

    def _message_properties(self):
        return [{'linkcode': self._linkcode},
                {'group': self._group},
                {'address': self._address},
                {'category': self._category},
                {'subcategory': self._subcategory},
                {'firmware': self._firmware}]




"""Sensor states."""
from enum import Enum
from insteonplm.constants import (COMMAND_LIGHT_ON_0X11_NONE,
                                  COMMAND_LIGHT_OFF_0X13_0X00,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01,
                                  MESSAGE_TYPE_BROADCAST_MESSAGE,
                                  MESSAGE_TYPE_ALL_LINK_CLEANUP,
                                  MESSAGE_TYPE_ALL_LINK_BROADCAST)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.states import State


class LeakSensorState(Enum):
    """Enum to define dry/wet state of the leak sensor."""
    DRY = 0
    WET = 1


class SensorBase(State):
    """Base state representing a non-controlable variable value sensor.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      register_updates(self, callback)
    """

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the State state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        template_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
            cmd2=None)
        template_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
            cmd2=None)

        template_on_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None)
        template_off_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None)

        self._message_callbacks.add(template_on_broadcast,
                                    self._sensor_on_command_received)
        self._message_callbacks.add(template_off_broadcast,
                                    self._sensor_off_command_received)

        self._message_callbacks.add(template_on_group,
                                    self._sensor_on_command_received)
        self._message_callbacks.add(template_off_group,
                                    self._sensor_off_command_received)

    def _sensor_on_command_received(self, msg):
        """Message handler for Standard or Extended sensor on messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x11 for on.
        """
        self._update_subscribers(msg.cmd2)

    def _sensor_off_command_received(self, msg):
        """Message handler for Standard or Extended sensor off messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x13 for off.
        """
        self._update_subscribers(0x00)


class VariableSensor(SensorBase):
    """Device state representing a variable value sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      register_updates(self, callback)
    """


class OnOffSensor(SensorBase):
    """Device state representing an On/Off sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def _sensor_on_command_received(self, msg):
        """Message handler for Standard or Extended sensor on messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x11 for on.
        """
        self._update_subscribers(0x01)


class SmokeCO2Sensor(SensorBase):
    """Device state representing a Smoke/CO2 sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the State state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        template_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None))
        template_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None))
        template_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None))

        self._message_callbacks.add(template_broadcast,
                                    self._sensor_state_received)
        self._message_callbacks.add(template_cleanup,
                                    self._sensor_state_received)
        self._message_callbacks.add(template_group,
                                    self._sensor_state_received)

    def _sensor_state_received(self, msg):
        self._update_subscribers(msg.targetHi)


class IoLincSensor(SensorBase):
    """Device state representing a I/O Linc Sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the IoLinkSensor."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._updatemethod = self._send_status_request

        template_open_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None))
        template_close_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
            cmd2=None)

        template_open_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None))
        template_close_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=None)

        template_open_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=self._group)
        template_close_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=self._group)

        self._message_callbacks.add(template_open_broadcast,
                                    self._open_message_received)
        self._message_callbacks.add(template_close_broadcast,
                                    self._close_message_received)

        self._message_callbacks.add(template_open_cleanup,
                                    self._open_message_received)
        self._message_callbacks.add(template_close_cleanup,
                                    self._close_message_received)

        self._message_callbacks.add(template_open_group,
                                    self._open_message_received)
        self._message_callbacks.add(template_close_group,
                                    self._close_message_received)

    def _send_status_request(self):
        status_command = StandardSend(self._address,
                                      COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01)
        self._send_method(status_command, self._status_message_received)

    def _open_message_received(self, msg):
        self._update_subscribers(0x01)

    def _close_message_received(self, msg):
        self._update_subscribers(0x00)

    def _status_message_received(self, msg):
        if msg.cmd2 == 0x00:
            self._update_subscribers(0x00)
        else:
            self._update_subscribers(0x01)


class LeakSensorDryWet(State):
    """Water leak sensor for the Dry or Wet states.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None,
                 dry_wet: LeakSensorState=None):
        """Initialize the LeakSensorDry state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._dry_wet_type = dry_wet
        self._dry_wet_callbacks = []

        template_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None))

        template_all_link = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None))

        template_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None))

        self._message_callbacks.add(template_broadcast,
                                    self._dry_wet_message_received)

        self._message_callbacks.add(template_all_link,
                                    self._dry_wet_message_received)

        self._message_callbacks.add(template_cleanup,
                                    self._dry_wet_message_received)

    def register_dry_wet_callback(self, callback):
        """Register the callback for the wet and dry state callbacks."""
        self._dry_wet_callbacks.append(callback)

    def set_value(self, dry_wet: LeakSensorState):
        """Set the value of the state to dry or wet."""
        value = 0
        if dry_wet == self._dry_wet_type:
            value = 1
        self._update_subscribers(value)

    def _dry_wet_message_received(self, msg):
        """The sensor is reporting a dry or a wet state."""
        for callback in self._dry_wet_callbacks:
            callback(self._dry_wet_type)
        self._update_subscribers(0x01)


class LeakSensorHeartbeat(State):
    """Water leak sensor for the Dry state.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the LeakSensorDry state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._dry_wet_callbacks = []

        template_dry_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None))

        template_wet_broadcast = StandardReceive.template(
            commandtuple={'cmd1': 0x13, 'cmd2': self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None))

        template_dry_all_link = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None))

        template_wet_all_link = StandardReceive.template(
            commandtuple={'cmd1': 0x13, 'cmd2': self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None))

        template_dry_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None))

        template_wet_cleanup = StandardReceive.template(
            commandtuple={'cmd1': 0x13, 'cmd2': self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None))

        self._message_callbacks.add(template_dry_broadcast,
                                    self._dry_message_received)
        self._message_callbacks.add(template_wet_broadcast,
                                    self._wet_message_received)
        self._message_callbacks.add(template_dry_all_link,
                                    self._dry_message_received)
        self._message_callbacks.add(template_wet_all_link,
                                    self._wet_message_received)
        self._message_callbacks.add(template_dry_cleanup,
                                    self._dry_message_received)
        self._message_callbacks.add(template_wet_cleanup,
                                    self._wet_message_received)

    def register_dry_wet_callback(self, callback):
        """Register the callback for the wet and dry state callbacks."""
        self._dry_wet_callbacks.append(callback)

    def set_value(self, dry_wet: LeakSensorState):
        """Set the state to wet or dry."""
        if dry_wet == LeakSensorState.DRY:
            self._update_subscribers(0x11)
        else:
            self._update_subscribers(0x13)

    def _dry_message_received(self, msg):
        """The sensor is reporting a dry state."""
        for callback in self._dry_wet_callbacks:
            callback(LeakSensorState.DRY)
        self._update_subscribers(0x11)

    def _wet_message_received(self, msg):
        """The sensor is reporting a wet state."""
        for callback in self._dry_wet_callbacks:
            callback(LeakSensorState.WET)
        self._update_subscribers(0x13)



"""Test message creation."""

import insteonplm.messages
from insteonplm.messages.allLinkComplete import AllLinkComplete
from insteonplm.messages.allLinkRecordResponse import (
    AllLinkRecordResponse)
from insteonplm.messages.buttonEventReport import (
    ButtonEventReport)
from insteonplm.messages.extendedReceive import ExtendedReceive
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.getIMInfo import GetImInfo
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.standardSend import StandardSend
from insteonplm.address import Address


def test_create_standardReceive_message():
    """Test create standardReceive message."""
    address1 = 0x11
    address2 = 0x22
    address3 = 0x33
    target1 = 0x44
    target2 = 0x55
    target3 = 0x66
    flags = 0x77
    cmd1 = 0x88
    cmd2 = 0x99
    rawmessage = bytearray([0x02, 0x50, address1, address2, address3, target1,
                            target2, target3, flags, cmd1, cmd2])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, StandardReceive)
    assert msg.address == Address(bytearray([address1, address2, address3]))
    assert msg.target == Address(bytearray([target1, target2, target3]))
    assert msg.cmd1 == cmd1
    assert msg.cmd2 == cmd2


def test_create_extendedReceive_message():
    """Test create extendedReceive message."""
    address1 = 0x01
    address2 = 0x02
    address3 = 0x03
    target1 = 0x04
    target2 = 0x05
    target3 = 0x06
    flags = 0x07
    cmd1 = 0x08
    cmd2 = 0x09
    rawmessage = bytearray([0x02, 0x51, address1, address2, address3, target1,
                            target2, target3, flags, cmd1, cmd2,
                            0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88,
                            0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, ExtendedReceive)
    assert msg.address == Address(bytearray([address1, address2, address3]))
    assert msg.target == Address(bytearray([target1, target2, target3]))
    assert msg.cmd1 == cmd1
    assert msg.cmd2 == cmd2


def test_create_allLinkComplete_message():
    """Test create allLinkComplete message."""
    linkcode = 0x11
    group = 0x22
    address1 = 0x33
    address2 = 0x44
    address3 = 0x55
    cat = 0x66
    subcat = 0x77
    firmware = 0x88

    rawmessage = bytearray([0x02, 0x53, linkcode, group, address1, address2,
                            address3, cat, subcat, firmware])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, AllLinkComplete)

    assert msg.linkcode == linkcode
    assert msg.group == group
    assert msg.address == Address(bytearray([address1, address2, address3]))
    assert msg.category == cat
    assert msg.subcategory == subcat
    assert msg.firmware == firmware


def test_button_event_report():
    """Test button event report."""
    event = 0x02
    rawmessage = bytearray([0x02, 0x54, event])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, ButtonEventReport)
    assert msg.event == event
    assert msg.eventText == 'SET button tapped'


def test_AllLinkRecordResponse_message():
    """Test AllLinkRecordResponse message."""
    flags = 0x11
    group = 0x22
    address1 = 0x33
    address2 = 0x44
    address3 = 0x55
    linkdata1 = 0x66
    linkdata2 = 0x77
    linkdata3 = 0x88
    rawmessage = bytearray([0x02, 0x57, flags, group, address1, address2,
                            address3, linkdata1, linkdata2, linkdata3])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, AllLinkRecordResponse)
    assert msg.group == group
    assert msg.address == Address(bytearray([address1, address2, address3]))
    assert msg.linkdata1 == linkdata1
    assert msg.linkdata2 == linkdata2
    assert msg.linkdata3 == linkdata3


def test_GetImInfo_message():
    """Test GetImInfo message."""
    address1 = 0x11
    address2 = 0x22
    address3 = 0x33
    cat = 0x44
    subcat = 0x55
    firmware = 0x66
    acknak = 0x77
    rawmessage = bytearray([0x02, 0x60, address1, address2, address3, cat,
                            subcat, firmware, acknak])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, GetImInfo)
    assert msg.address == Address(bytearray([address1, address2, address3]))
    assert msg.category == cat
    assert msg.subcategory == subcat
    assert msg.firmware == firmware


def test_StandardSend_withAcknak_message():
    """Test StandardSend withAcknak message."""
    target1 = 0x11
    target2 = 0x22
    target3 = 0x33
    flags = 0xEF  # 11101111
    cmd1 = 0x55
    cmd2 = 0x66
    acknak = 0x06
    rawmessage = bytearray([0x02, 0x62, target1, target2, target3, flags,
                            cmd1, cmd2, acknak])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, StandardSend)
    assert msg.cmd1 == cmd1
    assert msg.cmd2 == cmd2
    assert msg.isack
    assert not msg.isnak


def test_ExtendedSend_withAcknak_message():
    """Test ExtendedSend withAcknak message."""
    target1 = 0x11
    target2 = 0x22
    target3 = 0x33
    flags = 0x10  # 00010000
    cmd1 = 0x55
    cmd2 = 0x66
    acknak = 0x06
    rawmessage = bytearray(
        [0x02, 0x62, target1, target2, target3, flags, cmd1, cmd2,
         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, acknak])
    msg, buffer = insteonplm.messages.create(rawmessage)

    assert isinstance(msg, ExtendedSend)
    assert msg.cmd1 == cmd1
    assert msg.cmd2 == cmd2
    assert msg.isack
    assert not msg.isnak


def test_iscomplete_with_complete_message():
    """Test iscomplete with complete message."""
    rawmessage = bytearray([0x02, 0x50, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00, 0x00])
    assert insteonplm.messages.iscomplete(rawmessage)


def test_iscomplete_with_incomplete_message():
    """Test iscomplete with incomplete message."""
    shortmessage = bytearray([0x02, 0x50, 0x00])
    assert not insteonplm.messages.iscomplete(shortmessage)


def test_incomplete_standard_message():
    """Test incomplete standard message."""
    rawmessage = bytearray([0x02, 0x50, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00])
    msg, buffer = insteonplm.messages.create(rawmessage)
    assert msg is None


def test_incomplete_extended_message():
    """Test incomplete extended message."""
    rawmessage = bytearray([0x02, 0x62, 0x1a, 0x2b, 0x3c, 0x10, 0x7d, 0x8e,
                            0x9f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
    msg, buffer = insteonplm.messages.create(rawmessage)
    assert msg is None

    rawmessage.append(0x06)
    msg, buffer = insteonplm.messages.create(rawmessage)
    assert isinstance(msg, ExtendedSend)


def test_leading_unknown_messge():
    """Test leading unknown messge."""
    rawmessage = bytearray([0x02, 0x00, 0x15, 0x02, 0x50, 0x46, 0xd0, 0xe6,
                            0x43, 0x6c, 0x15, 0x40, 0x11, 0x01])
    msg, buffer = insteonplm.messages.create(rawmessage)
    assert isinstance(msg, StandardReceive)
    assert msg.cmd1 == 0x11
    assert msg.cmd2 == 0x01



"""INSTEON Device Unknown Device Type."""
from insteonplm.devices import Device, ALDB, ALDBVersion


class UnknownDevice(Device):
    """Unknown Device Class.

    Unknown Device used when only the device address is known but no other
    information. Available methods:
        id_request()
        product_data_request()
        assign_to_all_link_group()
        delete_from_all_link_group()
        device_text_string_request()
        enter_linking_mode()
        enter_unlinking_mode()
        get_engine_version()
        ping()
        read_aldb()
        write_aldb()
    """

    def __init__(self, plm, address, cat=None, subcat=None, product_key=0x00,
                 description='', model=''):
        """Initalize the UnknownDevice Class."""
        super().__init__(plm, address, cat, subcat, product_key,
                         description, model)
        self._aldb = ALDB(None, None, self._address, version=ALDBVersion.Null)



"""Test all message classes."""
import binascii

from insteonplm.address import Address
from insteonplm.messages.allLinkCleanupFailureReport import (
    AllLinkCleanupFailureReport)
from insteonplm.messages.allLinkCleanupStatusReport import (
    AllLinkCleanupStatusReport)
from insteonplm.messages.allLinkComplete import AllLinkComplete
from insteonplm.messages.allLinkRecordResponse import AllLinkRecordResponse
from insteonplm.messages.buttonEventReport import ButtonEventReport
from insteonplm.messages.cancelAllLinking import CancelAllLinking
from insteonplm.messages.extendedReceive import ExtendedReceive
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.getFirstAllLinkRecord import GetFirstAllLinkRecord
from insteonplm.messages.getImConfiguration import GetImConfiguration
from insteonplm.messages.getIMInfo import GetImInfo
from insteonplm.messages.getNextAllLinkRecord import GetNextAllLinkRecord
from insteonplm.messages.resetIM import ResetIM
from insteonplm.messages.sendAlllinkCommand import SendAllLinkCommand
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.startAllLinking import StartAllLinking
from insteonplm.messages.userReset import UserReset
from insteonplm.messages.x10received import X10Received
from insteonplm.messages.x10send import X10Send


def test_allLinkCleanupStatusReport():
    """Test AllLinkCleanupStatusReport."""
    msg = AllLinkCleanupStatusReport(0x11)
    assert msg.acknak == 0x11
    assert msg.hex == hexmsg(0x02, 0x58, 0x11)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_allLinkComplete():
    """Test AllLinkComplete."""
    linkcode = 0x11
    group = 0x22
    address = bytearray([0x33, 0x44, 0x55])
    cat = 0x66
    subcat = 0x77
    firmware = 0x88
    msg = AllLinkComplete(linkcode, group, address, cat, subcat, firmware)
    assert msg.address == Address(address)
    assert msg.linkcode == linkcode
    assert msg.group == group
    assert msg.category == cat
    assert msg.subcategory == subcat
    assert msg.hex == hexmsg(0x02, 0x53, 0x11, 0x22, 0x33,
                             0x44, 0x55, 0x66, 0x77, 0x88)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_allLinkFailureReport():
    """Test AllLinkFailureReport."""
    group = 0x11
    addr = bytearray([0x22, 0x33, 0x44])
    msg = AllLinkCleanupFailureReport(group, addr)

    assert msg.address == Address(addr)
    assert msg.group == group
    assert msg.hex == hexmsg(0x02, 0x56, 0x01, 0x11, 0x22, 0x33, 0x44)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_allLinkRecordResponse():
    """Test AllLinkRecordResponse."""
    flag = 0x11
    group = 0x22
    addr = bytearray([0x33, 0x44, 0x55])
    link1 = 0x66
    link2 = 0x77
    link3 = 0x88
    msg = AllLinkRecordResponse(flag, group, addr, link1, link2, link3)
    assert msg.address == Address(addr)
    assert msg.group == group
    assert msg.linkdata1 == link1
    assert msg.linkdata2 == link2
    assert msg.linkdata3 == link3
    assert msg.hex == hexmsg(0x02, 0x57, flag, group, 0x33, 0x44, 0x55,
                             link1, link2, link3)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_buttonEventReport():
    """Test ButtonEventReport."""
    event = 0x03
    msg = ButtonEventReport(event)
    assert msg.event == event
    assert msg.hex == hexmsg(0x02, 0x54, event)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_cancelAllLinking():
    """Test CancelAllLinking."""
    msg = CancelAllLinking()
    assert msg.hex == hexmsg(0x02, 0x65)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = CancelAllLinking(0x06)
    assert msg.hex == hexmsg(0x02, 0x65, 0x06)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = CancelAllLinking(0x15)
    assert msg.hex == hexmsg(0x02, 0x65, 0x15)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_extendedReceive():
    """Test ExtendedReceive."""
    address = bytearray([0x11, 0x22, 0x33])
    target = bytearray([0x44, 0x55, 0x66])
    flags = 0x77
    cmd1 = 0x88
    cmd2 = 0x99
    userdata = {}
    userdatatest = bytearray()

    for i in range(1, 15):
        key = 'd' + str(i)
        userdata.update({key: 0xe0})
        userdatatest.append(0xe0)

    msg = ExtendedReceive(address, target, {'cmd1': cmd1, 'cmd2': cmd2},
                          userdata, flags=flags)
    assert msg.hex == hexmsg(0x02, 0x51, Address(address), Address(target),
                             flags, cmd1, cmd2, userdatatest)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_extendedSend():
    """Test ExtendedSend."""
    address = bytearray([0x11, 0x22, 0x33])
    flags = 0x44 | 0x10
    cmd1 = 0x55
    cmd2 = 0x66
    userdata = {}
    ack = 0x06
    nak = 0x15

    for i in range(1, 15):
        key = 'd'+str(i)
        val = 0xe0 + i
        userdata.update({key: val})

    msg = ExtendedSend(address, {'cmd1': cmd1, 'cmd2': cmd2},
                       userdata, flags=flags)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address),
                             flags | 0x10, cmd1, cmd2, userdata)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = ExtendedSend(address, {'cmd1': cmd1, 'cmd2': cmd2},
                       userdata, flags=flags, acknak=ack)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address),
                             flags | 0x10, cmd1, cmd2, userdata, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = ExtendedSend(address, {'cmd1': cmd1, 'cmd2': cmd2},
                       userdata, flags=flags, acknak=nak)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address),
                             flags | 0x10, cmd1, cmd2, userdata, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_getFirstAllLinkRecord():
    """Test GetFirstAllLinkRecord."""
    ack = 0x06
    nak = 0x15
    msg = GetFirstAllLinkRecord()
    assert msg.hex == hexmsg(0x02, 0x69)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = GetFirstAllLinkRecord(ack)
    assert msg.hex == hexmsg(0x02, 0x69, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = GetFirstAllLinkRecord(nak)
    assert msg.hex == hexmsg(0x02, 0x69, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_getImConfiguration():
    """Test GetImConfiguration."""
    ack = 0x06
    nak = 0x15
    flags = 0x11
    msg = GetImConfiguration()
    assert msg.hex == hexmsg(0x02, 0x73)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = GetImConfiguration(flags, ack)
    assert msg.hex == hexmsg(0x02, 0x73, flags, 0x00, 0x00, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = GetImConfiguration(flags, nak)
    assert msg.hex == hexmsg(0x02, 0x73, flags, 0x00, 0x00, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_getIMInfo():
    """Test GetIMInfo."""
    addr = bytearray([0x11, 0x22, 0x33])
    cat = 0x44
    subcat = 0x55
    firmware = 0x66
    ack = 0x06
    nak = 0x15

    msg = GetImInfo()
    assert msg.hex == hexmsg(0x02, 0x60)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = GetImInfo(addr, cat, subcat, firmware, ack)
    assert msg.hex == hexmsg(0x02, 0x60, addr, cat, subcat, firmware, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = GetImInfo(addr, cat, subcat, firmware, nak)
    assert msg.hex == hexmsg(0x02, 0x60, addr, cat, subcat, firmware, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_getNextAllLinkRecord():
    """Test GetNextAllLinkRecord."""
    ack = 0x06
    nak = 0x15

    msg = GetNextAllLinkRecord()
    assert msg.hex == hexmsg(0x02, 0x6a)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = GetNextAllLinkRecord(ack)
    assert msg.hex == hexmsg(0x02, 0x6a, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = GetNextAllLinkRecord(nak)
    assert msg.hex == hexmsg(0x02, 0x6a, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_resetIM():
    """Test ResetIM."""
    ack = 0x06
    nak = 0x15

    msg = ResetIM()
    assert msg.hex == hexmsg(0x02, 0x67)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = ResetIM(ack)
    assert msg.hex == hexmsg(0x02, 0x67, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = ResetIM(nak)
    assert msg.hex == hexmsg(0x02, 0x67, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_sendAlllinkCommand():
    """Test SendAlllinkCommand."""
    group = 0x11
    cmd1 = 0x22
    cmd2 = 0x33
    ack = 0x06
    nak = 0x15

    msg = SendAllLinkCommand(group, cmd1, cmd2)
    assert msg.hex == hexmsg(0x02, 0x61, group, cmd1, cmd2)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = SendAllLinkCommand(group, cmd1, cmd2, ack)
    assert msg.hex == hexmsg(0x02, 0x61, group, cmd1, cmd2, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = SendAllLinkCommand(group, cmd1, cmd2, nak)
    assert msg.hex == hexmsg(0x02, 0x61, group, cmd1, cmd2, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_standardReceive():
    """Test StandardReceive."""
    address = bytearray([0x11, 0x22, 0x33])
    target = bytearray([0x44, 0x55, 0x66])
    flags = 0x77
    cmd1 = 0x88
    cmd2 = 0x99

    msg = StandardReceive(address, target, {'cmd1': cmd1, 'cmd2': cmd2},
                          flags=flags)
    assert msg.hex == hexmsg(0x02, 0x50, Address(address), Address(target),
                             flags, cmd1, cmd2)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_standardSend():
    """Test StandardSend."""
    address = bytearray([0x11, 0x22, 0x33])
    flags = 0x44
    cmd1 = 0x55
    cmd2 = 0x66
    ack = 0x06
    nak = 0x15

    msg = StandardSend(address, {'cmd1': cmd1, 'cmd2': cmd2}, flags=flags)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address), flags, cmd1, cmd2)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = StandardSend(address, {'cmd1': cmd1, 'cmd2': cmd2},
                       flags=flags, acknak=ack)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address),
                             flags, cmd1, cmd2, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = StandardSend(address, {'cmd1': cmd1, 'cmd2': cmd2},
                       flags=flags, acknak=nak)
    assert msg.hex == hexmsg(0x02, 0x62, Address(address),
                             flags, cmd1, cmd2, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_startAllLinking():
    """Test StartAllLinking."""
    group = 0x11
    code = 0x03
    ack = 0x06
    nak = 0x15

    msg = StartAllLinking(code, group)
    assert msg.hex == hexmsg(0x02, 0x64, code, group)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = StartAllLinking(code, group, ack)
    assert msg.hex == hexmsg(0x02, 0x64, code, group, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = StartAllLinking(code, group, nak)
    assert msg.hex == hexmsg(0x02, 0x64, code, group, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def test_userReset():
    """Test UserReset."""
    msg = UserReset()
    assert msg.hex == hexmsg(0x02, 0x55)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_x10received():
    """Test X10Received."""
    rawX10 = 0x11
    flag = 0x22
    msg = X10Received(rawX10, flag)
    assert msg.hex == hexmsg(0x02, 0x52, rawX10, flag)
    assert len(msg.hex)/2 == msg.sendSize
    assert len(msg.hex)/2 == msg.receivedSize


def test_x10send():
    """Test X10Send."""
    rawX10 = 0x11
    flag = 0x22
    ack = 0x06
    nak = 0x15

    msg = X10Send(rawX10, flag)
    assert msg.hex == hexmsg(0x02, 0x63, rawX10, flag)
    assert not msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.sendSize

    msg = X10Send(rawX10, flag, ack)
    assert msg.hex == hexmsg(0x02, 0x63, rawX10, flag, ack)
    assert msg.isack
    assert not msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize

    msg = X10Send(rawX10, flag, nak)
    assert msg.hex == hexmsg(0x02, 0x63, rawX10, flag, nak)
    assert not msg.isack
    assert msg.isnak
    assert len(msg.hex)/2 == msg.receivedSize


def hexmsg(*arg):
    """Utility method to convert a message to hex."""
    msg = bytearray([])
    for b in arg:
        if b is None:
            pass
        elif isinstance(b, int):
            msg.append(b)
        elif isinstance(b, Address):
            msg.extend(b.bytes)
        elif isinstance(b, bytearray):
            msg.extend(b)
        elif isinstance(b, bytes):
            msg.extend(b)
        elif isinstance(b, dict):
            for i in range(1, 15):
                key = 'd' + str(i)
                val = b[key]
                msg.append(val)

    return binascii.hexlify(msg).decode()



"""INSTEON Device Type Dimmable Lighting Control Module."""
from insteonplm.devices import Device
from insteonplm.states.dimmable import DimmableSwitch, DimmableSwitch_Fan


class DimmableLightingControl(Device):
    """Dimmable Lighting Controller.

    INSTEON On/Off switch device class. Available device control options are:
        - light_on(onlevel=0xff)
        - light_on_fast(onlevel=0xff)
        - light_off()
        - light_off_fast()

    To monitor changes to the state of the device subscribe to the state
    monitor:
         - lightOnLevel.connect(callback)  (state='LightOnLevel')

    where callback defined as:
        - callback(self, device_id, state, state_value)
    """

    def __init__(self, plm, address, cat, subcat, product_key=None,
                 description=None, model=None):
        """Initialize the DimmableLightingControl Class."""
        Device.__init__(self, plm, address, cat, subcat, product_key,
                        description, model)

        self._stateList[0x01] = DimmableSwitch(
            self._address, "lightOnLevel", 0x01, self._send_msg,
            self._message_callbacks, 0x00)


class DimmableLightingControl_2475F(DimmableLightingControl):
    """FanLinc model 2475F Dimmable Lighting Control.

    Device Class 0x01 subcat 0x2e

    Two separate INSTEON On/Off switch devices are created with ID
        1) Ligth
            - ID: xxxxxx (where xxxxxx is the Insteon address of the device)
            - Controls:
                - light_on(onlevel=0xff)
                - light_on_fast(onlevel=0xff)
                - light_off()
                - light_off_fast()
            - Monitor: lightOnLevel.connect(callback)
        2) Fan
            - ID: xxxxxx_2  (where xxxxxx is the Insteon address of the device)
            - Controls:
                - fan_on(onlevel=0xff)
                - fan_off()
                - light_on(onlevel=0xff)  - Same as fan_on(onlevel=0xff)
                - light_off()  - Same as fan_off()
            - Monitor: fanSpeed.connect(callback)

    where callback defined as:
        - callback(self, device_id, state, state_value)
    """

    def __init__(self, plm, address, cat, subcat, product_key=None,
                 description=None, model=None):
        """Initalize the DimmableLightingControl_2475F Class."""
        super().__init__(plm, address, cat, subcat, product_key,
                         description, model)

        self._stateList[0x01] = DimmableSwitch(
            self._address, "lightOnLevel", 0x01, self._send_msg,
            self._message_callbacks, 0x00)
        self._stateList[0x02] = DimmableSwitch_Fan(
            self._address, "fanOnLevel", 0x02, self._send_msg,
            self._message_callbacks, 0x00)



"""Test insteonplm.evices module."""
import asyncio
import logging

from insteonplm.constants import (COMMAND_LIGHT_OFF_0X13_0X00,
                                  COMMAND_LIGHT_ON_0X11_NONE,
                                  MESSAGE_ACK)
from insteonplm.messages.standardSend import StandardSend
import insteonplm.devices
from insteonplm.devices.dimmableLightingControl import DimmableLightingControl
from tests.mockPLM import MockPLM
from insteonplm.devices import DIRECT_ACK_WAIT_TIMEOUT

_LOGGING = logging.getLogger(__name__)


def test_create_device():
    """Test create device."""
    plm = MockPLM()
    device = insteonplm.devices.create(plm, '112233', 0x01, 0x0d, None)
    assert device.id == '112233'
    assert isinstance(device, DimmableLightingControl)


def test_create_device_from_bytearray():
    """Test create device from byte array."""
    plm = MockPLM()
    target = bytearray()
    target.append(0x01)
    target.append(0x0d)
    device = insteonplm.devices.create(plm, '112233',
                                       target[0], target[1],
                                       None)
    assert device.id == '112233'
    assert isinstance(device, DimmableLightingControl)


def test_send_msg():

    @asyncio.coroutine
    def run_test(loop):
        mockPLM = MockPLM(loop)
        address = '1a2b3c'
        device = insteonplm.devices.create(mockPLM, address, 0x01, 0x0d, 0x44)
        mockPLM.devices[address] = device

        # Send the ON command. This should be sent directly to the PLM
        device.states[0x01].on()
        yield from asyncio.sleep(.1, loop=loop)

        # Send the OFF command. This should wait in queue until the
        # Direct ACK timeout
        device.states[0x01].off()
        yield from asyncio.sleep(.1, loop=loop)

        # ACK the ON command
        msgreceived = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE,
                                   cmd2=0xff, flags=0x00, acknak=MESSAGE_ACK)
        mockPLM.message_received(msgreceived)
        asyncio.sleep(.1, loop=loop)

        _LOGGING.debug('Assert that the ON command is the command in the PLM')
        assert mockPLM.sentmessage == StandardSend(
            address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff, flags=0x00).hex

        # Sleep until the Direct ACK time out should expire
        yield from asyncio.sleep(DIRECT_ACK_WAIT_TIMEOUT + .2,
                                 loop=loop)

        # Confirm that the OFF command made it to the PLM
        assert mockPLM.sentmessage == StandardSend(
            address, COMMAND_LIGHT_OFF_0X13_0X00).hex

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))



"""Test Dimmable Lighting Control devices."""

import asyncio
from insteonplm.constants import (COMMAND_LIGHT_OFF_0X13_0X00,
                                  COMMAND_LIGHT_ON_0X11_NONE,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_NONE,
                                  MESSAGE_ACK,
                                  MESSAGE_TYPE_ALL_LINK_BROADCAST,
                                  MESSAGE_TYPE_DIRECT_MESSAGE_ACK)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.devices.dimmableLightingControl import (
    DimmableLightingControl, DimmableLightingControl_2475F)
from .mockPLM import MockPLM
from .mockCallbacks import MockCallbacks


def test_dimmableLightingControl():
    """Test generic Dimmable Lighting Control devices."""
    def run_test(loop):
        """Asyncio test to run."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'
        cat = 0x01
        subcat = 0x04
        product_key = None
        description = 'SwitchLinc Dimmer (1000W)'
        model = '2476DH'
        device = DimmableLightingControl(plm, address, cat, subcat,
                                         product_key, description, model)

        assert device.address.hex == address
        assert device.cat == cat
        assert device.subcat == subcat
        assert device.product_key == 0x00  # Product key should not be None
        assert device.description == description
        assert device.model == model
        assert device.id == address

        plm.devices[device.address.hex] = device

        callbacks = MockCallbacks()
        device.states[0x01].register_updates(callbacks.callbackmethod1)

        device.states[0x01].on()
        yield from asyncio.sleep(.1, loop=loop)
        plm.message_received(StandardSend(
            address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
            acknak=MESSAGE_ACK))
        yield from asyncio.sleep(.1, loop=loop)
        plm.message_received(StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
            flags=MessageFlags.create(
                MESSAGE_TYPE_DIRECT_MESSAGE_ACK, 0, 2, 3)))
        yield from asyncio.sleep(.1, loop=loop)
        sentmsg = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0xff

        device.states[0x01].off()
        yield from asyncio.sleep(.1, loop=loop)
        recievedmsg = StandardSend(address, COMMAND_LIGHT_OFF_0X13_0X00,
                                   acknak=MESSAGE_ACK)
        plm.message_received(recievedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        recievedmsg = StandardReceive(
            address, target, COMMAND_LIGHT_OFF_0X13_0X00,
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        plm.message_received(recievedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        sentmsg = StandardSend(address, COMMAND_LIGHT_OFF_0X13_0X00)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0x00

        device.states[0x01].set_level(0x55)
        yield from asyncio.sleep(.1, loop=loop)
        recievedmsg = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE,
                                   cmd2=0x55, acknak=MESSAGE_ACK)
        plm.message_received(recievedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        recievedmsg = StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x55,
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        plm.message_received(recievedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        sentmsg = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE,
                               cmd2=0x55)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0x55

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_dimmableLightingControl_manual_changes():
    """Test manual changes to Dimmable Lighting Controls."""
    def run_test(loop):
        """Asyncio test method."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        cat = 0x01
        subcat = 0x04
        product_key = None
        description = 'SwitchLinc Dimmer (1000W)'
        model = '2476DH'
        device = DimmableLightingControl(plm, address, cat, subcat,
                                         product_key, description, model)

        assert device.address.hex == address
        assert device.cat == cat
        assert device.subcat == subcat
        assert device.product_key == 0x00  # Product key should not be None
        assert device.description == description
        assert device.model == model
        assert device.id == address

        plm.devices[device.address.hex] = device

        callbacks = MockCallbacks()
        device.states[0x01].register_updates(callbacks.callbackmethod1)

        receivedmsg = StandardReceive(
            address, '000001', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x66,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST,
                                      0, 2, 3))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0x66

        receivedmsg = StandardReceive(
            address, '000001', COMMAND_LIGHT_OFF_0X13_0X00,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST,
                                      0, 2, 3))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0x00

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_dimmableLightingControl_status():
    """Test status updates for Dimmable Lighting Controls."""
    def run_test(loop):
        """Asyncio test to run."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'
        cat = 0x01
        subcat = 0x04
        product_key = None
        description = 'SwitchLinc Dimmer (1000W)'
        model = '2476DH'

        device = DimmableLightingControl(plm, address, cat, subcat,
                                         product_key, description, model)

        assert device.address.hex == address
        assert device.cat == cat
        assert device.subcat == subcat
        assert device.product_key == 0x00  # Product key should not be None
        assert device.description == description
        assert device.model == model
        assert device.id == address
        plm.devices[device.address.hex] = device

        callbacks = MockCallbacks()
        device.states[0x01].register_updates(callbacks.callbackmethod1)

        device.async_refresh_state()
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardSend(
            address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
            acknak=MESSAGE_ACK)
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardReceive(
            address, target, {'cmd1': 0x08, 'cmd2': 0x27},
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        sentmsg = StandardSend(address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0x27

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_switchedLightingControl_2475F():
    """Test device 2475F."""
    def run_test(loop):
        """Asyncio test."""
        class fanLincStatus(object):
            """Callback class to capture sensor changes."""

            lightOnLevel = None
            fanOnLevel = None

            def device_status_callback1(self, device_id, state, value):
                """Callback method to capture light changes."""
                self.lightOnLevel = value

            def device_status_callback2(self, device_id, state, value):
                """Callback method to capture fan changes."""
                self.fanOnLevel = value

        mockPLM = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'

        cat = 0x01
        subcat = 0x2e
        product_key = 0x00
        description = 'FanLinc Dual Band'
        model = '2475F'

        callbacks = fanLincStatus()

        device = DimmableLightingControl_2475F(mockPLM, address, cat, subcat,
                                               product_key, description, model)
        
        mockPLM.devices[device.address.hex] = device
        
        assert device.states[0x01].name == 'lightOnLevel'
        assert device.states[0x02].name == 'fanOnLevel'

        device.states[0x01].register_updates(callbacks.device_status_callback1)
        device.states[0x02].register_updates(callbacks.device_status_callback2)

        device.states[0x01].async_refresh_state()
        yield from asyncio.sleep(.1, loop=loop)
        ackmsg = StandardSend(address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                              acknak=MESSAGE_ACK)
        statusmsg = StandardReceive(
            address, target, {'cmd1': 0xdf, 'cmd2': 0x55},
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        mockPLM.message_received(ackmsg)
        yield from asyncio.sleep(.1, loop=loop)
        mockPLM.message_received(statusmsg)
        yield from asyncio.sleep(.1, loop=loop)

        assert callbacks.lightOnLevel == 0x55

        device.states[0x02].async_refresh_state()
        yield from asyncio.sleep(.1, loop=loop)
        ackmsg = StandardSend(address, {'cmd1': 0x19, 'cmd2': 0x03},
                              flags=0x00, acknak=MESSAGE_ACK)
        statusmsg = StandardReceive(
            address, target, {'cmd1': 0xab, 'cmd2': 0x77},
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        mockPLM.message_received(ackmsg)
        yield from asyncio.sleep(.1, loop=loop)
        mockPLM.message_received(statusmsg)
        yield from asyncio.sleep(.1, loop=loop)

        assert callbacks.lightOnLevel == 0x55
        assert callbacks.fanOnLevel == 0x77

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_dimmableLightingControl_2475F_status():
    """Test device 2475F status updates."""
    def run_test(loop):
        """Asyncio test."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'

        cat = 0x01
        subcat = 0x2e
        product_key = 0x00
        description = 'FanLinc Dual Band'
        model = '2475F'

        device = DimmableLightingControl_2475F(plm, address, cat, subcat,
                                               product_key, description, model)
        
        plm.devices[device.address.hex] = device
        callbacks = MockCallbacks()
        device.states[0x02].register_updates(callbacks.callbackmethod1)

        device.states[0x02].async_refresh_state()
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardSend(
            address, COMMAND_LIGHT_STATUS_REQUEST_0X19_NONE, cmd2=0x03,
            acknak=MESSAGE_ACK)
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        receivedmsg = StandardReceive(
            address, target, {'cmd1': 0x08, 'cmd2': 0x27},
            flags=MessageFlags.create(MESSAGE_TYPE_DIRECT_MESSAGE_ACK,
                                      0, 2, 3))
        plm.message_received(receivedmsg)
        yield from asyncio.sleep(.1, loop=loop)
        sentmsg = StandardSend(
            address, COMMAND_LIGHT_STATUS_REQUEST_0X19_NONE, cmd2=0x03)
        assert plm.sentmessage == sentmsg.hex
        assert callbacks.callbackvalue1 == 0x27

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))



"""Insteon Powerline Modem Interface Module.

This module provides a unified asyncio network handler for interacting with
Insteon Powerline modems like the 2413U and 2412S.
"""
import aiohttp
import asyncio
import binascii
import logging
import serial
import serial.aio

from insteonplm.plm import PLM, Hub

__all__ = ('Connection')
_LOGGER = logging.getLogger(__name__)

# pylint: disable=invalid-name,no-member
try:
    ensure_future = asyncio.ensure_future
except AttributeError:
    ensure_future = asyncio.async


@asyncio.coroutine
def create_http_connection(loop, protocol_factory, host, port=25105,
                           auth=None, connector=None):
    session = aiohttp.ClientSession(auth=auth, connector=connector)
    protocol = protocol_factory()
    transport = HttpTransport(loop, protocol, session, host, port)
    return (transport, protocol)


class Connection:
    """Handler to maintain the Powerline device connection."""

    @classmethod
    @asyncio.coroutine
    def create(cls, device='/dev/ttyUSB0', host=None,
               username=None, password=None, port=25010,
               auto_reconnect=True, loop=None, workdir=None,
               poll_devices=True):
        """Initiate a connection to a specific device.

        Here is where we supply the device and callback callables we
        expect for this PLM class object.

        :param device:
            Unix device where the PLM is attached
        :param address:
            IP Address of the Hub
        :param username:
            User name for connecting to the Hub
        :param password:
            Password for connecting to the Hub
        :param auto_reconnect:
            Should the Connection try to automatically reconnect if needed?
        :param loop:
            asyncio.loop for async operation

        :type device:
            str
        :type auto_reconnect:
            boolean
        :type loop:
            asyncio.loop
        :type update_callback:
            callable
        """
        conn = cls()

        conn.device = device
        conn.host = host
        conn.username = username
        conn.password = password
        conn.port = port
        conn._loop = loop or asyncio.get_event_loop()
        conn._retry_interval = 1
        conn._closed = False
        conn._closing = False
        conn._halted = False
        conn._auto_reconnect = auto_reconnect

        def connection_lost():
            """Function callback for Protocol when connection is lost."""
            if conn._auto_reconnect and not conn._closing:
                ensure_future(conn._reconnect(), loop=conn._loop)

        protocol_class = PLM
        if conn.host:
            protocol_class = Hub
        conn.protocol = protocol_class(
            connection_lost_callback=connection_lost,
            loop=conn._loop,
            workdir=workdir,
            poll_devices=poll_devices)

        yield from conn._reconnect()

        return conn

    @property
    def transport(self):
        """Return pointer to the transport object.

        Use this property to obtain passthrough access to the underlying
        Protocol properties and methods.
        """
        return self.protocol.transport

    def _get_retry_interval(self):
        return self._retry_interval

    def _reset_retry_interval(self):
        self._retry_interval = 1

    def _increase_retry_interval(self):
        self._retry_interval = min(300, 1.5 * self._retry_interval)

    # TODO:
    # This code need to change to handle serial or HTTP connections.
    @asyncio.coroutine
    def _reconnect(self):
        while True:
            try:
                if self._halted:
                    yield from asyncio.sleep(2, loop=self._loop)
                else:
                    if self.host:
                        _LOGGER.info('Connecting to Hub on %s', self.host)
                        hub_loop = asyncio.new_event_loop()
                        auth = aiohttp.BasicAuth(self.username, self.password)
                        connector = aiohttp.TCPConnector(
                            limit=1, loop=self._loop, keepalive_timeout=10)
                        yield from create_http_connection(
                            self._loop, lambda: self.protocol,
                            self.host, port=self.port,
                            auth=auth, connector=connector)
                    else:
                        _LOGGER.info('Connecting to PLM on %s', self.device)
                        yield from serial.aio.create_serial_connection(
                            self._loop, lambda: self.protocol,
                            self.device, baudrate=19200)
                        self._reset_retry_interval()
                    return

            except OSError:
                self._increase_retry_interval()
                interval = self._get_retry_interval()
                _LOGGER.warning('Connecting failed, retry in %i seconds: %s',
                                 interval, self.device)
                yield from asyncio.sleep(interval, loop=self._loop)

    def close(self):
        """Close the PLM device connection and don't try to reconnect."""
        _LOGGER.warning('Closing connection to PLM')
        self._closing = True
        self._auto_reconnect = False
        if self.protocol.transport:
            self.protocol.transport.close()

    def halt(self):
        """Close the PLM device connection and wait for a resume() request."""
        _LOGGER.warning('Halting connection to PLM')
        self._halted = True
        if self.protocol.transport:
            self.protocol.transport.close()

    def resume(self):
        """Resume the PLM device connection if we have been halted."""
        _LOGGER.warning('Resuming connection to PLM')
        self._halted = False

    @property
    def dump_conndata(self):
        """Developer tool for debugging forensics."""
        attrs = vars(self)
        return ', '.join("%s: %s" % item for item in attrs.items())


class HttpTransport(asyncio.Transport):
    """An asyncio transport model of an HTTP communication channel.

    A transport class is an abstraction of a communication channel.
    This allows protocol implementations to be developed against the
    transport abstraction without needing to know the details of the
    underlying channel, such as whether it is a pipe, a socket, or
    indeed an HTTP connection.


    You generally won’t instantiate a transport yourself; instead, you
    will call `create_http_connection` which will create the
    transport and try to initiate the underlying communication channel,
    calling you back when it succeeds.
    """

    def __init__(self, loop, protocol, session, host, port=25105):
        """Initialize the HttpTransport class."""
        super().__init__()
        self._loop = loop
        self._protocol = protocol
        self._session = session
        self._host = host
        self._port = port

        self._closing = False
        self._protocol_paused = False
        self._max_read_size = 1024
        self._write_buffer = []
        self._has_reader = False
        self._has_writer = False
        self._poll_wait_time = 0.0005
        self._read_write_lock = asyncio.Lock(loop=self._loop)
        self._last_read = asyncio.Queue(loop=self._loop)

        # TODO: restart _ensure_reader if it fails or if the connection
        # is lost.
        asyncio.ensure_future(self._ensure_reader(), loop=self._loop)
        

    def abort(self):
        self._session.close()

    def can_write_eof(self):
        return False

    def close(self):
        _LOGGER.debug("Closing session")
        self._loop.run_until_complete(self._close())
        _LOGGER.info("Hub Session closed")

    @asyncio.coroutine
    def _close(self):
        yield from self._session.close()
        yield from asyncio.sleep(0, loop=self._loop)
        self._closing = True

    def get_write_buffer_size(self):
        return 0

    def pause_reading(self):
        self._loop.call_soon(self._lock_read, True)

    def resume_reading(self):
        self._loop.call_soon(self._lock_read, False)

    def set_write_buffer_limits(self, high=None, low=None):
        raise NotImplementedError(
            "HTTP connections do not support write buffer limits")

    def write(self, data):
        _LOGGER.debug("..................Writing a message..............")
        hex_data = binascii.hexlify(data).decode()
        url = 'http://{:s}:{:d}/3?{:s}=I=3'.format(self._host, self._port,
                                                   hex_data)
        coro_write = self._async_write(url)
        asyncio.ensure_future(coro_write, loop=self._loop)

    @asyncio.coroutine
    def _async_write(self, url):
        _LOGGER.debug("Writing message: %s", url)
        yield from self._read_write_lock
        resp = yield from self._session.post(url)
        self._write_last_read(0)
        if self._read_write_lock.locked():
            self._read_write_lock.release()
        _LOGGER.debug("Post status: %s", resp.status)
        return resp.status

    def write_eof(self):
        raise NotImplementedError(
            "HTTP connections do not support end-of-file")

    def writelines(self, list_of_data):
        raise NotImplementedError(
            "HTTP connections do not support writelines")

    def clear_buffer(self):
        """Clear the Hub read buffer."""
        asyncio.wait(self._clear_buffer())

    @asyncio.coroutine
    def _clear_buffer(self):
        _LOGGER.debug("..................Clearing the buffer..............")
        url = 'http://{:s}:{:d}/1?XB=M=1'.format(self._host, self._port)
        yield from self._async_write(url)

    @asyncio.coroutine
    def _ensure_reader(self):
        yield from self._clear_buffer()
        self._write_last_read(0)
        url = 'http://{:s}:{:d}/buffstatus.xml'.format(self._host, self._port)
        self._loop.call_soon(self._protocol.connection_made(self))
        while True:
            if self._session.closed:
                return
            yield from self._read_write_lock
            last_stop = 0
            if not self._last_read.empty():
                last_stop = self._last_read.get_nowait()
            response = yield from self._session.get(url)
            buffer = None
            if response.status == 200:
                html = yield from response.text()
                last_stop, buffer = self._parse_buffer(html, last_stop)
                self._write_last_read(last_stop)
            # TODO: handle status codes
            if self._read_write_lock.locked():
                self._read_write_lock.release()
            if buffer:
                _LOGGER.debug('New buffer: %s', buffer)
                bin_buffer = binascii.unhexlify(buffer)
                self._protocol.data_received(bin_buffer)
            yield from asyncio.sleep(1, loop=self._loop)

    def _parse_buffer(self, html, last_stop):
        buffer = ''
        raw_text = html.replace('<response><BS>', '')
        raw_text = raw_text.replace('</BS></response>', '')
        raw_text = raw_text.strip()
        if raw_text[:200] == '0'*200:
            # Likely the buffer was cleared
            return (0, None)
        this_stop = int(raw_text[-2:], 16)
        if this_stop > last_stop:
            _LOGGER.debug('Buffer from %d to %d', last_stop, this_stop)
            buffer = raw_text[last_stop:this_stop]
        elif this_stop < last_stop:
            _LOGGER.debug('Buffer from %d to 200 and 0 to %d',
                          last_stop, this_stop)
            buffer_hi = raw_text[last_stop:200]
            if buffer_hi == '0'*len(buffer_hi):
                #The buffer was probably reset since the last read
                buffer_hi = ''
            buffer_low = raw_text[0:this_stop]
            buffer = '{:s}{:s}'.format(buffer_hi, buffer_low)
        else:
            buffer = None
        return (this_stop, buffer)

    def _write_last_read(self, val):
        while not self._last_read.empty():
            not_used = self._last_read.get_nowait()
        self._last_read.put_nowait(val)



"""Test insteonplm Address type."""
from insteonplm.address import Address


def test_textstring():
    """Test Address created from text string."""
    addr = Address('1a2b3c')    
    assert addr.human == '1A.2B.3C'
    assert addr.hex == '1a2b3c'
    assert addr.bytes == b'\x1a\x2b\x3c'


def test_bytearray():
    """Test Address created from bytearray."""
    addr_ba = bytearray([0x1a, 0x2b, 0x3c])
    addr = Address(addr_ba)
    assert addr.human == '1A.2B.3C'
    assert addr.hex == '1a2b3c'
    assert addr.bytes == b'\x1a\x2b\x3c'


def test_bytes():
    """Test Address created from bytes string."""
    addr_b = b'\x1a\x2b\x3c'
    addr = Address(addr_b)
    assert addr.human == '1A.2B.3C'
    assert addr.hex == '1a2b3c'
    assert addr.bytes == b'\x1a\x2b\x3c'


def test_none():
    """Test Address equal to None."""
    addr = Address(None)
    assert addr.human == '00.00.00'
    assert addr.hex == '000000'
    assert addr.bytes == b'\x00\x00\x00'


def test_eq():
    """Test Addresses pattern matching."""
    addr1 = Address(None)
    addr2 = Address('1a2b3c')
    addr3 = Address('4d5e6f')

    assert addr1.matches_pattern(addr2)
    assert addr3.matches_pattern(addr1)
    assert not (addr2.matches_pattern(addr3))
    assert addr2.matches_pattern(addr2)

def test_x10():
    addr = Address.x10('A', 5)

    assert addr.hex == '000601'
    assert addr.human == 'X10.A.05'
    assert addr.is_x10
    assert addr.x10_housecode == 'A'
    assert addr.x10_unitcode == 5
    assert addr.x10_housecode_byte == 6
    assert addr.x10_unitcode_byte == 1

    addr2 = Address.x10('A', 20)
    assert addr2.human == 'X10.A.20'


"""INSTEON Message Send All-Link Command."""
from insteonplm.constants import (MESSAGE_SEND_ALL_LINK_COMMAND_0X61,
                                  MESSAGE_SEND_ALL_LINK_COMMAND_SIZE,
                                  MESSAGE_SEND_ALL_LINK_COMMAND_RECEIVED_SIZE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.messages.message import Message


class SendAllLinkCommand(Message):
    """Insteon Send All Link Command Message.

    Message type 0x6A
    """

    _code = MESSAGE_SEND_ALL_LINK_COMMAND_0X61
    _sendSize = MESSAGE_SEND_ALL_LINK_COMMAND_SIZE
    _receivedSize = MESSAGE_SEND_ALL_LINK_COMMAND_RECEIVED_SIZE
    _description = 'Insteon Get Next All Link Record Message'

    def __init__(self, group, allLinkCommand, broadcastCommand, acknak=None):
        """Initalize the SendAllLinkCommand Class."""
        self._group = group
        self._allLinkCommmand = allLinkCommand
        self._broadcastCommand = broadcastCommand

        self._acknak = self._setacknak(acknak)

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return SendAllLinkCommand(rawmessage[2],
                                  rawmessage[3],
                                  rawmessage[4],
                                  rawmessage[5:6])

    @property
    def group(self):
        """Return the All-Link group."""
        return self._group

    @property
    def allLinkCommmand(self):
        """Return the All-Link command."""
        return self._allLinkCommmand

    @property
    def broadcastCommand(self):
        """Return the broadcase command."""
        return self._broadcastCommand

    @property
    def isack(self):
        """Test if this is an ACK message."""
        return self._acknak is not None and self._acknak == MESSAGE_ACK

    @property
    def isnak(self):
        """Test if this is a NAK message."""
        return self._acknak is not None and self._acknak == MESSAGE_NAK

    def _message_properties(self):
        return [{'group': self._group},
                {'allLinkCommand': self._allLinkCommmand},
                {'broadcastCommand': self._broadcastCommand},
                {'acknak': self._acknak}]



3#!/usr/bin/env python
"""Setup for insteonplm module."""
from setuptools import setup, find_packages

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    print('Skipping md->rst conversion for long_description')
    long_description = 'Error converting Markdown from git repo'

if len(long_description) < 100:
    print("\n***\n***\nWARNING: %s\n***\n***\n" % long_description)

setup(
    name='insteonplm',
    version='0.11.2',
    author='David McNett',
    author_email='nugget@macnugget.org',
    url='https://github.com/nugget/python-insteonplm',
    license="LICENSE",
    packages=find_packages(),
    scripts=[],
    description='Python API for controlling Insteon PowerLinc Modems',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'pyserial==3.2.0',
        'pyserial-asyncio',
        'async_timeout'
    ],
    entry_points={
        'console_scripts': ['insteonplm_monitor = insteonplm.tools:monitor',
                            'insteonplm_interactive = '
                            'insteonplm.tools:interactive']
    }
)



"""Test the INSTEON PLM device."""
import asyncio
import async_timeout
import logging
import binascii

import insteonplm
from insteonplm.constants import (COMMAND_LIGHT_ON_0X11_NONE,
                                  COMMAND_LIGHT_OFF_0X13_0X00,
                                  MESSAGE_FLAG_BROADCAST_0X80,
                                  COMMAND_ID_REQUEST_0X10_0X00,
                                  COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                                  MESSAGE_NAK,
                                  MESSAGE_ACK,
                                  X10_COMMAND_ON,
                                  X10_COMMAND_OFF)
from insteonplm.plm import PLM
from insteonplm.address import Address
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.getIMInfo import GetImInfo
from insteonplm.messages.getFirstAllLinkRecord import GetFirstAllLinkRecord
from insteonplm.messages.getNextAllLinkRecord import GetNextAllLinkRecord
from insteonplm.messages.allLinkRecordResponse import AllLinkRecordResponse
from insteonplm.messages.x10received import X10Received

from .mockConnection import MockConnection
from .mockCallbacks import MockCallbacks

_LOGGER = logging.getLogger()
SEND_MSG_WAIT = 1.1
SEND_MSG_ACKNAK_WAIT = .2
DIRECT_ACK_WAIT_TIMEOUT = 3.1
RECV_MSG_WAIT = .1


@asyncio.coroutine
def wait_for_plm_command(plm, cmd, loop):
    try:
        with async_timeout.timeout(10, loop=loop):
            while not plm.transport.lastmessage == cmd.hex:
                yield from asyncio.sleep(.1, loop=loop)
            _LOGGER.info('Expected message sent %s', cmd)
            return True
    except asyncio.TimeoutError:
        _LOGGER.error('Expected message not sent %s', cmd)
        return False


@asyncio.coroutine
def do_plm(loop):
    """Asyncio coroutine to test the PLM."""
    _LOGGER.info('Connecting to Insteon PLM')

    # pylint: disable=not-an-iterable
    conn = yield from MockConnection.create(loop=loop)

    def async_insteonplm_light_callback(device):
        """Log that our new device callback worked."""
        _LOGGER.warn('New Device: %s %02x %02x %s, %s', device.id, device.cat,
                     device.subcat, device.description, device.model)

    def async_light_on_level_callback(device_id, state, value):
        """Callback to catch changes to light value."""
        _LOGGER.info('Device %s state %s value is changed to %02x',
                     device_id, state, value)

    conn.protocol.add_device_callback(async_insteonplm_light_callback)

    plm = conn.protocol
    plm.connection_made(conn.transport)

    _LOGGER.info('Replying with IM Info')
    _LOGGER.info('_____________________')
    cmd_sent = yield from wait_for_plm_command(plm, GetImInfo(), loop)
    if not cmd_sent:
        assert False
    msg = insteonplm.messages.getIMInfo.GetImInfo(address='1a2b3c', cat=0x03,
                                                  subcat=0x20, firmware=0x00,
                                                  acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)
    assert plm.address == Address('1a2b3c')
    assert plm.cat == 0x03
    assert plm.subcat == 0x20
    assert plm.product_key == 0x00

    _LOGGER.info('Replying with an All-Link Record')
    _LOGGER.info('________________________________')
    cmd_sent = yield from wait_for_plm_command(plm, GetFirstAllLinkRecord(),
                                               loop)
    if not cmd_sent:
        assert False
    msg = GetFirstAllLinkRecord(MESSAGE_ACK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    msg = AllLinkRecordResponse(
        flags=0x00, group=0x01, address='4d5e6f',
        linkdata1=0x01, linkdata2=0x0b, linkdata3=0x000050)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    _LOGGER.info('Replying with Last All-Link Record')
    _LOGGER.info('__________________________________')
    cmd_sent = yield from wait_for_plm_command(plm, GetNextAllLinkRecord(),
                                               loop)
    if not cmd_sent:
        assert False
    msg = GetNextAllLinkRecord(MESSAGE_NAK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    _LOGGER.info('Replying with Device Info Record')
    _LOGGER.info('________________________________')
    msg = StandardSend('4d5e6f', COMMAND_ID_REQUEST_0X10_0X00)
    cmd_sent = wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False
    msg = StandardSend('4d5e6f', COMMAND_ID_REQUEST_0X10_0X00,
                       acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT, loop=loop)
    msg = StandardReceive(
        address='4d5e6f', target='010b00',
        commandtuple={'cmd1': 0x01, 'cmd2': 0x00},
        flags=MESSAGE_FLAG_BROADCAST_0X80)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)
    for addr in plm.devices:
        _LOGGER.info('Device: %s', addr)

    _LOGGER.info('Replying with Device Status Record')
    _LOGGER.info('__________________________________')
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
                       acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    msg = insteonplm.messages.standardReceive.StandardReceive(
        address='4d5e6f', target='1a2b3c',
        commandtuple={'cmd1': 0x17, 'cmd2': 0xaa},
        flags=0x20)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    assert plm.devices['4d5e6f'].states[0x01].value == 0xaa

    _LOGGER.info('Testing device message ACK/NAK handling')
    _LOGGER.info('_______________________________________')
    plm.devices['4d5e6f'].states[0x01].on()
    plm.devices['4d5e6f'].states[0x01].off()
    plm.devices['4d5e6f'].states[0x01].set_level(0xbb)

    # Test that the first ON command is sent
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False

    # ACK the ON command
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_ON_0X11_NONE,
                       cmd2=0xff, flags=0x00, acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)

    # Let the Direct ACK timeout expire
    yield from asyncio.sleep(DIRECT_ACK_WAIT_TIMEOUT)

    # Test that the Off command has now sent
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_OFF_0X13_0X00)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False

    # NAK the OFF command and test that the OFF command is resent
    plm.transport.lastmessage = ''
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_OFF_0X13_0X00,
                       acknak=MESSAGE_NAK)
    plm.data_received(msg.bytes)

    msg = StandardSend('4d5e6f', COMMAND_LIGHT_OFF_0X13_0X00)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False

    # Let the ACK/NAK timeout expire and test that the OFF command is resent
    plm.transport.lastmessage = ''
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_OFF_0X13_0X00)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False

    # ACK the OFF command and let the Direct ACK message expire
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_OFF_0X13_0X00,
                       acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)

    # Test that the second SET_LEVEL command is sent
    msg = StandardSend('4d5e6f', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xbb)
    cmd_sent = yield from wait_for_plm_command(plm, msg, loop)
    if not cmd_sent:
        assert False


@asyncio.coroutine
def do_plm_x10(loop):
    """Asyncio coroutine to test the PLM X10 message handling."""
    _LOGGER.info('Connecting to Insteon PLM')

    # pylint: disable=not-an-iterable
    conn = yield from MockConnection.create(loop=loop)
    
    cb = MockCallbacks()

    #conn.protocol.add_device_callback(async_insteonplm_light_callback)

    plm = conn.protocol
    plm.connection_made(conn.transport)

    _LOGGER.info('Replying with IM Info')
    _LOGGER.info('_____________________')
    cmd_sent = yield from wait_for_plm_command(plm, GetImInfo(), loop)
    if not cmd_sent:
        assert False
    msg = insteonplm.messages.getIMInfo.GetImInfo(address='1a2b3c', cat=0x03,
                                                  subcat=0x20, firmware=0x00,
                                                  acknak=MESSAGE_ACK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    _LOGGER.info('Replying with an All-Link Record NAK')
    _LOGGER.info('____________________________________')
    cmd_sent = yield from wait_for_plm_command(plm, GetFirstAllLinkRecord(),
                                               loop)
    if not cmd_sent:
        assert False
    msg = GetFirstAllLinkRecord(MESSAGE_NAK)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(RECV_MSG_WAIT)

    _LOGGER.info('Add X10 device and receive messages')
    _LOGGER.info('____________________________________')
    housecode = 'C'
    unitcode = 9
    plm.add_x10_device(housecode, unitcode, 'OnOff')
    id = Address.x10(housecode, unitcode).id
    plm.devices[id].states[0x01].register_updates(cb.callbackmethod1)

    msg = X10Received.unit_code_msg(housecode, unitcode)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(.1, loop=loop)
    msg = X10Received.command_msg(housecode, X10_COMMAND_ON)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(.1, loop=loop)
    assert cb.callbackvalue1 == 0xff

    msg = X10Received.unit_code_msg(housecode, unitcode)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(.1, loop=loop)
    msg = X10Received.command_msg(housecode, X10_COMMAND_OFF)
    plm.data_received(msg.bytes)
    yield from asyncio.sleep(.1, loop=loop)
    assert cb.callbackvalue1 == 0x00


def test_plm():
    """Main test for the PLM."""
    logging.basicConfig(level=logging.DEBUG)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(do_plm(loop))


def test_plm_x10():
    """Test X10 message handling."""
    logging.basicConfig(level=logging.DEBUG)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(do_plm_x10(loop))
